/* up:  array.man */

#include <stdio.h>
#include <spa.h>


#define AarrayType 1234

struct aarray {
    int len;
    int alloc;
    int eltsize;
    char *contents;
};

int TmpAREFi;
#define aref(a, i) \
((void *)(TmpAREFi = (i), \
	  TmpAREFi < 0 || TmpAREFi >= (a)->len \
	  ? (char *) error("aref: Out of bounds") \
	  : (a)->contents + TmpAREFi * (a)->eltsize))

extern int new_entry_p;
void aget1 (void *xp, struct aarray *a, int i);
void aput1 (void *xp, struct aarray *a, int i);
int apush1 (void *xp, struct aarray *a);
int atop1 (void *xp, struct aarray *a);
int apop1 (void *xp, struct aarray *a);
void *aelt (struct aarray *a, int i);
int aid (struct aarray *a, char *x);    /* the header file says void *x (on purpose) */
void *aneed (struct aarray *a, int i);
void *apull (struct aarray *a);
void *aallocate (struct aarray *a, int len);
void azero (struct aarray *a);
void agrow (struct aarray *a, int newalloc);
void asetlength (struct aarray *a, int newlen);
void agrowlength (struct aarray *a, int newlen);
struct aarray *make_aarray (int alloc, int eltsize);
void free_aarray (struct aarray *a);
void write_aarray (struct aarray *a, FILE *file);
struct aarray *read_aarray (FILE *file);
void dump_aarray (struct aarray *a, int (*print_elt)(), FILE *file);
void save_aarray (struct aarray *a, char *filename);
struct aarray *load_aarray (char *filename);
void save_as_array (struct aarray *a, char *filename);
struct aarray *load_as_aarray (char *filename, int alloc, int eltsize);



/**  GET_ELT() computes the address of the i-th element, and signals
***  an error if it's out of bounds.  
**/

#define GET_ELT(p, a, i, where) \
  if (i < 0 || i >= a->len) \
    error("%s: Out of bounds: %d", where, i); \
  p = a->contents + i * a->eltsize



/**-------------------------------------------------------------------
***  Access by value
**/

/**  aget() copies the i-th element into the given location.  A typical
***  call looks like aget(&x, a, i), where x is a variable of the same
***  type as the elements in the array.  
**/

void aget1 (void *xp, struct aarray *a, int i)
{
    void *p;
    GET_ELT(p, a, i, "aget");
    bcopy(p, xp, a->eltsize);
}


/**  aput() copies from a variable into an array element.  
**/

void aput1 (void *xp, struct aarray *a, int i)
{
    void *p;
    GET_ELT(p, a, i, "aput");
    bcopy(xp, p, a->eltsize);
}


/**  Copy the variable whose address is given into the element at the
***  end of the vector, increasing the vector length by one.
***
***  x may be NULL, in which case the length of the vector is
***  increased by one, but no copy is done.
***
***  Returns the index of the new element (the old length).  
**/

int apush1 (void *xp, struct aarray *a)
{
    int i;
    void *p;

    i = a->len;
    asetlength(a, i+1);
    GET_ELT(p, a, i, "apush");
    if (xp != NULL)
      bcopy(xp, p, a->eltsize);
    return i;
}

/**  atop() copies the last element in the array into the variable
***  whose address is given.  It returns the index of the copied element,
***  or -1 if the array is empty.
***
***  If x is NULL, no copy is done, but the return value is still
***  useful.  
**/

int atop1 (void *xp, struct aarray *a)
{
    void *p;
    int i;

    i = a->len - 1;
    if (i >= 0 && xp != NULL) {
	GET_ELT(p, a, i, "atop");
	bcopy(p, xp, a->eltsize);
    }
    return i;
}


/**  apop() fetches the topmost element, then shrinks the vector by one.
***  Its return value is the index of the fetched element, or -1 if the
***  vector is empty.
***
***  The target pointer may be NULL, in which case the vector is shrunk
***  and the index (new length) returned, but no fetch is done.
**/

int apop1 (void *xp, struct aarray *a)
{
    int i;
    void *p;

    if (a->len <= 0) return -1;
    a->len--;
    i = a->len;
    if (xp != NULL) {
	bcopy(a->contents + i * a->eltsize, xp, a->eltsize);
    }
    bzero(a->contents + a->len * a->eltsize, a->eltsize);
    return i;
}


/**-------------------------------------------------------------------
***  Access by address
**/

/**  aelt(a,i) returns the address of the ith element in a.
**/

void *aelt (struct aarray *a, int i)
{
    void *p;
    GET_ELT(p, a, i, "elt");
    return (void *) p;
}


/**  aid() computes i, given the address of the i-th element.  
**/

int aid (struct aarray *a, char *x)
{
    char *start = a->contents;
    char *end = a->contents + a->len * a->eltsize;
    
    if (x < start || x >= end)
      error("aid: Address out of bounds: 0x%x", x);
    return (x - start) / a->eltsize;
}

void *aneed (struct aarray *a, int i)
{
    if (i < a->len) {
	new_entry_p = 0;
	return aref(a, i);
    }
    else {
	agrowlength(a, i+1);
	new_entry_p = 1;
	return aref(a, i);
    }
}

/**  vpull() increases the length of the vector by one, and puts
***  the _address_ of the new element into the provided variable.
***  NOTE BENE: vpull() is the only function that behaves like this.
***  All the others take a pointer to the type of element stored
***  in the array.  vpull() takes a pointer to a pointer.  The reason
***  is that there would be no point in vpull copying the new element
***  itself into the variable, since the contents of the new element
***  is guaranteed to be zero.
***
***  The return value is the index of the new element.  
**/

void *apull (struct aarray *a)
{
    int i;
    void *p;

    i = a->len;
    asetlength(a, i+1);
    GET_ELT(p, a, i, "apull");
    return p;
}

void *aallocate (struct aarray *a, int len)
{
    int i;
    void *p;

    i = a->len;
    asetlength(a, i+len);
    GET_ELT(p, a, i, "aallocate1");
    return p;
}


/**-------------------------------------------------------------------
***  Length and allocation
**/

/**  Zero out an array.  Only the addresses within the current length
***  of the vector need to be zeroed out.  Addresses beyond the end of
***  the array are guaranteed to contain zeros.  
**/

void azero (struct aarray *a)
{
    if (a->len > 0) {
	bzero(a->contents, a->len * a->eltsize);
	a->len = 0;
    }
}

void azero_to (int len, struct aarray *a)
{
  if (a->len > len) {
    bzero(a->contents + len * a->eltsize, (a->len - len) * a->eltsize);
    a->len = len;
  }
}

/**
***  agrow() increases the allocation of an array.  The allocation is 
***  increased by the smallest multiple of a->realloc that makes the array
***  long enough.
***
***  asetlength() changes the length of an array.  If necessary, it also
***  grows the array.
***
***  agrowlength() is like asetlength(), except it won't decrease the length.
***
***  A newly created array is guaranteed to contain all zeros.  Hence, we
***  do not need to zero out addresses when we grow an array.  However, if
***  we do not zero out when we grow an array, then we need to zero out when
***  we shrink, so that we can rely on addresses in the allocation but beyond
***  the end of the array containing zeros.  
**/

void agrow (struct aarray *a, int newalloc)
{
    int oldbytealloc, newbytealloc;

    if (newalloc > a->alloc) {
	newalloc = newalloc * 4 / 3;
	oldbytealloc = a->alloc * a->eltsize;
	newbytealloc = newalloc * a->eltsize;
	reallocate(a->contents, oldbytealloc, newbytealloc, char);
	a->alloc = newalloc;
    }
}


void asetlength (struct aarray *a, int newlen)
{
    if (newlen > a->len) {
	agrow(a, newlen);
	a->len = newlen;
    }
    else if (newlen < a->len) {
	bzero(a->contents + newlen * a->eltsize,
	      (a->len - newlen) * a->eltsize);
	a->len = newlen;
    }
}


void agrowlength (struct aarray *a, int newlen)
{
    if (newlen > a->len) {
	agrow(a, newlen);
	a->len = newlen;
    }
}


/**-------------------------------------------------------------------
***  Create, Read, Write, Free
**/


struct aarray *make_aarray (int alloc, int eltsize)
{
    struct aarray *a;

    if (eltsize <= 0) error("make_aarray: Bad eltsize %d", eltsize);
    allocate(a, 1, struct aarray);
    a->len = 0;
    a->alloc = alloc;
    a->eltsize = eltsize;
    allocate(a->contents, alloc * eltsize, char);
    return a;
}


void free_aarray (struct aarray *a)
{
    free(a->contents);
    free(a);
}


void write_aarray (struct aarray *a, FILE *file)
{
    write_int(AarrayType, file);
    write_int(a->len, file);
    write_int(a->alloc, file);
    write_int(a->eltsize, file);
    write_array(a->contents, a->len * a->eltsize, char, file);
}


struct aarray *read_aarray (FILE *file)
{
    struct aarray *a;
    int x, size;

    if (read_int(x, file) == EOF) return NULL;
    if (x != AarrayType) error("read_aarray: Not an aarray");
    allocate(a, 1, struct aarray);
    read_int(a->len, file);
    read_int(a->alloc, file);
    read_int(a->eltsize, file);
    allocate(a->contents, a->alloc * a->eltsize, char);
    read_array(a->contents, a->len * a->eltsize, char, file);
    return a;
}


void dump_aarray (struct aarray *a, int (*print_elt)(), FILE *file)
{
    int i;
    
    fprintf(file, "Aarray *%d\n", a);
    fprintf(file, "len =       %d\n", a->len);
    fprintf(file, "alloc =     %d\n", a->alloc);
    fprintf(file, "eltsize =   %d\n", a->eltsize);
    fprintf(file, "contents =  *%d\n", a->contents);
    if (print_elt != NULL) {
	for (i = 0; i < a->len; i++) {
	    fprintf(file, "%-4d: ", i);
	    print_elt(aelt(a, i), file);
	    terpri(file);
	}
    }
}


void save_aarray (struct aarray *a, char *filename)
{
    FILE *file = Fopen(filename, "w");
    write_aarray(a, file);
    fclose(file);
}


struct aarray *load_aarray (char *filename)
{
    FILE *file = Fopen(filename, "r");
    struct aarray *a = read_aarray(file);
    fclose(file);
    return a;
}


void save_as_array (struct aarray *a, char *filename)
{
    FILE *file = Fopen(filename, "w");
    write_array(a->contents, a->len * a->eltsize, char, file);
    fclose(file);
}


struct aarray *load_as_aarray (char *filename, int alloc, int eltsize)
{
    if (file_exists_p(filename)) {
	int filesize = file_size(filename), len;
	struct aarray *a;
	FILE *file;
    
	if (filesize % eltsize != 0)
	  error("load_as_aarray: Size not a multiple of %d: %s", eltsize, filename);
	len = filesize / eltsize;
	a = make_aarray(max(alloc, len), eltsize);
	a->len = len;
	file = Fopen(filename, "r");
	read_array(a->contents, filesize, char, file);
	fclose(file);
	return a;
    }
    else return make_aarray(alloc, eltsize);
}
