#include <stdio.h>
#include "spa.h"
#include "aarray.h"
#include "rbtree.h"
#include "dict.h"


/**
***  See man amatrix-imp(3)
**/



/**************************************************************************/
/**                                                                      **/
/**  Submatrices excluding the root matrix are in one-one                **/
/**  correspondence with rbtree nodes.  The array `entries' is indexed   **/
/**  by rbtree node id's, which is to say, submat id's.  (Since rbtree   **/
/**  node id's start at 2, we can use 1 as the submat id for the root    **/
/**  matrix.)                                                            **/
/**                                                                      **/
/**  For a nontrivial (i.e., non-cell) submat i, Entries[i] contains     **/
/**  the table id of the submat's own component table (an rbtree node    **/
/**  id).  For a cell i, Entries[i] contains the value-id of the         **/
/**  cell's value.                                                       **/
/**                                                                      **/
/**  Cell values are stored in the array values.  Value-id's are         **/
/**  indices into the Values array.                                      **/
/**                                                                      **/
/**  new_value() creates a new value location and returns its address.   **/
/**                                                                      **/
/**  LastValue is the value-id of the most recently created value.       **/
/**                                                                      **/
/**  value_address(v) returns the address of the value with the given    **/
/**  id.                                                                 **/
/**                                                                      **/
/**************************************************************************/

#define MAXNDIMS 10

typedef int SubmatId;
typedef int RowId;
typedef int TableId;
typedef int ValueId;

struct amatrix {
    int ndims;
    int eltsize;
    TableId root;
    struct tstore *submats;
    struct aarray *entries;
    struct aarray *values;
};    

struct amatrix *AMatrix;
static int AmTMPi;
static struct amatrix *AmTMPm;

#define NDims        (AMatrix->ndims)
#define Root         (AMatrix->root)
#define Submats      (AMatrix->submats)
#define Entries      (AMatrix->entries)
#define Values       (AMatrix->values)


#define new_value()       (apull(AMatrix->values))
#define LastValue         (AMatrix->values->len)
#define value_address(v)  (AMatrix->values == NULL ? &(v) : aref(AMatrix->values,(v)-1))


/**************************************************************************/
/**                                                                      **/
/**  Mat(i) returns the submat union for a SubmatId.                     **/
/**                                                                      **/
/**  need_mat(i) also returns a submat union, but it creates a new       **/
/**  entry if necessary.  Nasty things happen if you use Mat(i) when     **/
/**  submatrix i does not already exist.                                 **/
/**                                                                      **/
/**  am_table(i) is the table-id of the table representing submatrix     **/
/**  i.                                                                  **/
/**                                                                      **/
/**  am_value(i) is the value-id of cell i.  (i is a submat id.)         **/
/**                                                                      **/
/**************************************************************************/

struct mat1 {
    TableId tab;
};

struct cell1 {
    ValueId val;
};

union submat {
    struct mat1 mat;
    struct cell1 cell;
};

#define submat_obj(i) \
  (AmTMPi = (i), \
   AmTMPi == 0 ? NULL : \
   AmTMPi == 1 ? (union submat *) &Root : \
   (union submat *) aref(Entries, AmTMPi-2))

#define am_table(i)  (submat_obj(i)->mat.tab)
#define am_value(i)  (submat_obj(i)->cell.val)
    

/**************************************************************************/
/**                                                                      **/
/**  need_component(k,i) returns the submat id of the component with     **/
/**  key k in the submatrix i.  It creates a new component if            **/
/**  necessary.                                                          **/
/**                                                                      **/
/**  find_component(k,i) does not create a new component.                **/
/**                                                                      **/
/**  submat_length(i,m) returns the length of submatrix i viewed as a    **/
/**  vector of components.                                               **/
/**                                                                      **/
/**  submat_nonempty_components(i,m) returns the number of nonempty      **/
/**  components.                                                         **/
/**                                                                      **/
/**************************************************************************/

SubmatId need_component (int k, SubmatId i)
{
    i = Tinsert(k, &am_table(i), Submats);
    aneed(Entries, i-2);
    return i;
}

#define find_component(k,i)  (Tsearch((k), am_table(i), Submats))

int submat_length (int i, struct amatrix *m)
{
    int k,r;

    r = am_table(i);
    Tlist_last(&k, r, m->submats);
    return k+1;
}

int submat_nonempty_components (int i, struct amatrix *m)
{
    return Tlist_length(am_table(i), m->submats);
}
    


/**************************************************************************/
/**                                                                      **/
/**  am_insert(keytuple, m) returns the address of the value for the     **/
/**  given key-tuple, creating a new entry if necessary.                 **/
/**                                                                      **/
/**  submat_id(keytuple, n, m) returns the submat id corresponding to    **/
/**  the given key-tuple.                                                **/
/**                                                                      **/
/**************************************************************************/

void *am_insert (int *key, struct amatrix *m)
{
    int *cell;
    int i, d;
    struct amatrix *prev = AMatrix;
    void *vp;

    AMatrix = m;
    i = 1;
    for (d = 0; d < NDims; d++) {
	i = need_component(key[d], i);
    }
    if (m->eltsize == sizeof(int)) vp = &am_value(i);
    else {
	if (am_value(i) == 0) {
	    vp = new_value();
	    am_value(i) = LastValue;
	}
	else vp = value_address(am_value(i));
    }
    AMatrix = prev;
    return vp;
}

SubmatId submat_id (int *key, int n)
{
    int i, d;

    i = 1;
    for (d = 0; d < n; d++) {
	i = find_component(key[d], i);
	if (i == 0) break;
    }
    return i;
}

void *am_search (int *keytuple, struct amatrix *m)
{
    int *cell;
    int i, d;
    struct amatrix *prev = AMatrix;
    void *vp;

    AMatrix = m;
    i = 1;
    for (d = 0; d < NDims; d++) {
	i = find_component(keytuple[d], i);
	if (i == 0) return NULL;
    }
    if (m->eltsize == sizeof(int)) vp = &am_value(i);
    else {
	if (am_value(i) == 0) error("am_search: cell without a value location");
	else vp = value_address(am_value(i));
    }
    AMatrix = prev;
    return vp;
}


/**************************************************************************/
/**                                                                      **/
/**  Iterating over components.  Basically just iteration over a         **/
/**  tlist.                                                              **/
/**                                                                      **/
/**  start_components (ComponentList *, SubmatId)                        **/
/**                                                                      **/
/**  SubmatId next_component (int *, ComponentList *)                    **/
/**                                                                      **/
/**  The row given to am_start_row must be left alone until you are      **/
/**  done iterating through it.                                          **/
/**                                                                      **/
/**************************************************************************/

typedef struct tree_state ComponentList;

#define start_components(iter,i) \
  (Tlist_start((iter), am_table(i), Submats))

#define next_component(kp,iter) \
  ((SubmatId) Tlist_next((kp), (struct tree_state *)(iter), Submats))
    
struct am_row {
    SubmatId i;
    int keytuple[MAXNDIMS];
    struct amatrix *m;
};

struct am_cell {
    int key;
    void *value;
};

struct am_cells {
    struct am_row *row;
    ComponentList components;
};




int am_get_row (struct am_row *row, int *keytuple, struct amatrix *m)
{
    struct amatrix *prev = AMatrix;
    SubmatId i;

    AMatrix = m;
    i = submat_id(keytuple, m->ndims - 1);
    if (i == 0) {
	AMatrix = prev;
	return 0;
    }
    row->i = i;
    bcopy(keytuple, row->keytuple, (m->ndims-1)*sizeof(int));
    row->m = m;
    AMatrix = prev;
    return 1;
}

am_start_row (struct am_cells *iter, struct am_row *row)
{
    struct amatrix *prev = AMatrix;
    
    AMatrix = row->m;
    iter->row = row;
    start_components(&iter->components, row->i);
    AMatrix = prev;
}

int am_next_cell (struct am_cell *cell, struct am_cells *iter)
{
    SubmatId i;
    struct amatrix *prev = AMatrix;
    struct am_row *row = iter->row;
    
    AMatrix = row->m;
    i = next_component(&cell->key, &iter->components);
    if (i == 0) return 0;
    cell->value = value_address(am_value(i));
    return 1;
}


/**************************************************************************/
/**                                                                      **/
/**  Recursing through all submatrices.                                  **/
/**                                                                      **/
/**  d0 is the depth in the matrix where we start, where depth is the    **/
/**  same as keytuple length.  (The root matrix has depth 0.)  dN is     **/
/**  the depth at which to truncate search: we do not go through the     **/
/**  components of any matrix at depth dN.  d is the depth of the        **/
/**  matrix whose components we are currently iterating through.         **/
/**  &components[d] is the component-list for the submat at depth d.     **/
/**  keytuples contains the keytuple for the matrix currently being      **/
/**  visited, and n is the length.                                       **/
/**                                                                      **/
/**  Note that the matrix at which we start is NOT included in the       **/
/**  iteration.  The first call to am_next_submat returns the first      **/
/**  component of the matrix where we start.                             **/
/**                                                                      **/
/**************************************************************************/

struct am_rows {
    struct amatrix *mat;
    int keytuple[MAXNDIMS];
    int n;
    int d0;
    int dN;
    ComponentList components[MAXNDIMS];
};

int am_start_submats (struct am_rows *iter, int targdim, int *keytuple, int n,
		      struct amatrix *m)
{
    SubmatId i;
    struct amatrix *prev = AMatrix;
    
    AMatrix = m;
    iter->mat = m;
    if (n > 0) bcopy(keytuple, iter->keytuple, n*sizeof(int));
    iter->d0 = n;
    iter->n = n;
    iter->dN = NDims - targdim;

    if (n >= iter->dN) iter->n = -1;
    else {
	i = submat_id(keytuple, n);
	if (i == 0) iter->n = -1;
	else start_components(&iter->components[n], i);
    }
    AMatrix = prev;
}

SubmatId am_next_submat (struct am_rows *iter)
{
    struct amatrix *prev = AMatrix;
    SubmatId i;
    int d = iter->n;

    if (d < iter->d0) return 0;

    AMatrix = iter->mat;
    if (d < iter->dN) i = next_component(&iter->keytuple[d], &iter->components[d]);
    else i = 0;
    while (i == 0) {
	d = --iter->n;
	if (d < iter->d0) {
	    AMatrix = prev;
	    return 0;
	}
	i = next_component(&iter->keytuple[d], &iter->components[d]);
    }
    d = ++iter->n;
    if (d < iter->dN) start_components(&iter->components[d], i);
    AMatrix = prev;
    return i;
}

SubmatId am_next_bottom_submat (struct am_rows *iter)
{
    SubmatId i;

    do i = am_next_submat(iter);
    while (i && iter->n != iter->dN);
    return i;
}


#define am_start_rows(iter,m)  (am_start_submats((iter), 1, NULL, 0, (m)))

am_next_row (struct am_row *row, struct am_rows *iter)
{
    SubmatId i;

    row->i = am_next_bottom_submat(iter);
    if (row->i == 0) return 0;
    bcopy(iter->keytuple, row->keytuple, (iter->mat->ndims-1)*sizeof(int));
    row->m = iter->mat;
    return 1;
}
    


/**===================================================================
***
***			   Make, Dump, Etc.
**/

struct amatrix *make_amatrix (int ndims, int alloc, int eltsize)
{
    struct amatrix *m;

    allocate(m, 1, struct amatrix);
    m->ndims = ndims;
    m->eltsize = eltsize;
    m->submats = make_tstore(alloc);
    m->entries = make_aarray(alloc, sizeof(int));
    if (eltsize > 0 && eltsize != sizeof(int))
      m->values = make_aarray(alloc, eltsize);
    return m;
}
    
int write_amatrix (struct amatrix *m, FILE *file)
{
    write_int(m->ndims, file);
    write_int(m->eltsize, file);
    write_int(m->root, file);
    write_tstore(m->submats, file);
    write_aarray(m->entries, file);
    if (m->values == 0) write_char(0, file);
    else {
	write_char(1, file);
	write_aarray(m->values, file);
    }
}

struct amatrix *read_amatrix (FILE *file)
{
    struct amatrix *m;

    allocate(m, 1, struct amatrix);
    read_int(m->ndims, file);
    read_int(m->eltsize, file);
    read_int(m->root, file);
    m->submats = read_tstore(file);
    m->entries = read_aarray(file);
    if (read_char(file)) {
	m->values = read_aarray(file);
    }
    return m;
}

int dump_amatrix (struct amatrix *m, int (*print_value)(), FILE *file)
{
    int d, i;
    
    fprintf(file, "AMatrix\n");
    fprintf(file, "Root = %d\n", m->root);
    terpri(file);
    fprintf(file, "Submats: (nodeid = submatid)\n");
    dump_tstore(m->submats, file);
    terpri(file);
    fprintf(file, "Entries:\n");
    fprintf(file, "submat entry\n");
    for (i = 0; i < m->entries->len; i++) {
	fprintf(file, "%6d %d\n", i+2, *(int*)aref(m->entries, i));
    }
    terpri(file);
    if (m->values == NULL) fprintf(file, "Values: NULL\n");
    else {
	fprintf(file, "Values:\n");
	fprintf(file, " id value\n");
	for (i = 0; i < m->values->len; i++) {
	    fprintf(file, "%3d ", i+1);
	    print_value(aref(m->values, i), file);
	}
    }
}


print_amatrix (struct amatrix *m, int (*print_elt)(), FILE *file)
{
    int i;
    struct am_row row;
    struct am_cell cell;
    struct am_rows rows;
    struct am_cells cells;
    
    am_start_rows(&rows, m);
    while (am_next_row(&row, &rows)) {
	for (i = 0; i < m->ndims-1; i++) {
	    if (i != 0) print_char(' ', file);
	    fprintf(file, "%d", row.keytuple[i]);
	}
	fprintf(file, ":\n");
	am_start_row(&cells, &row);
	while (am_next_cell(&cell, &cells)) {
	    fprintf(file, "  %d\t", cell.key);
	    print_elt(cell.value, file);
	    terpri(file);
	}
    }
}
