#include <spa.h>


struct amgr {
    int len;
    int alloc;
    int eltsize;
    void **loc;
};

extern int new_entry_p;

#define glength(a) ((a)->len)

#define BASE1(a)  ((char *) *(a)->loc)
#define ELT(a,i) (BASE1(a) + (i) * (a)->eltsize)
#define CHKID(a, i, where) \
  ((i) < 1 || (i) > (a)->len ? error("%s: ID out of bounds: %d", where, i) : 0)



/**-------------------------------------------------------------------
***  Access by value
**/


/**  Copy the variable whose address is given into the element at the
***  end of the vector, increasing the vector length by one.
***
***  x may be NULL, in which case the length of the vector is
***  increased by one, but no copy is done.
***
***  Returns the index of the new element (the old length).  
**/

int gpush (void *xp, struct amgr *a)
{
    int i;

    i = a->len + 1;
    gsetlength(a, i);
    if (xp != NULL) {
	bcopy(xp, ELT(a, i), a->eltsize);
    }
    return i;
}


/**  apop() fetches the topmost element, then shrinks the vector by one.
***  Its return value is the index of the fetched element, or -1 if the
***  vector is empty.
***
***  The target pointer may be NULL, in which case the vector is shrunk
***  and the index (new length) returned, but no fetch is done.
**/

int gpop (void *xp, struct amgr *a)
{
    int i;
    char *p;
    
    if (a->len <= 0) return 0;
    i = a->len;
    a->len--;
    p = ELT(a, i);
    if (xp != NULL) {
	bcopy(p, xp, a->eltsize);
    }
    bzero(p, a->eltsize);
    return i;
}


/**-------------------------------------------------------------------
***  Access by address
**/

int gneed (struct amgr *a, int i)
{
    if (i <= a->len) {
	new_entry_p = 0;
    }
    else {
	new_entry_p = 1;
	gsetlength(a, i);
    }
    return i;
}


/**  vpull() increases the length of the vector by one, and puts
***  the _address_ of the new element into the provided variable.
***  NOTE BENE: vpull() is the only function that behaves like this.
***  All the others take a pointer to the type of element stored
***  in the array.  vpull() takes a pointer to a pointer.  The reason
***  is that there would be no point in vpull copying the new element
***  itself into the variable, since the contents of the new element
***  is guaranteed to be zero.
***
***  The return value is the index of the new element.  
**/

int gnew (struct amgr *a)
{
    gsetlength(a, a->len+1);
    return a->len;
}

int galloc (struct amgr *a, int len)
{
    int i;
    void *p;

    i = a->len + 1;
    gsetlength(a, a->len + len);
    return i;
}


/**-------------------------------------------------------------------
***  Length and allocation
**/

/**  Zero out an array.  Only the addresses within the current length
***  of the vector need to be zeroed out.  Addresses beyond the end of
***  the array are guaranteed to contain zeros.  
**/

int gzero (struct amgr *a)
{
    if (a->len > 0) {
	char *p = (char *) *a->loc + a->eltsize;
	bzero(p, a->len * a->eltsize);
	a->len = 0;
    }
    return 1;
}

/**
***  agrow() increases the allocation of an array.  The allocation is 
***  increased by the smallest multiple of a->realloc that makes the array
***  long enough.
***
***  asetlength() changes the length of an array.  If necessary, it also
***  grows the array.
***
***  agrowlength() is like asetlength(), except it won't decrease the length.
***
***  A newly created array is guaranteed to contain all zeros.  Hence, we
***  do not need to zero out addresses when we grow an array.  However, if
***  we do not zero out when we grow an array, then we need to zero out when
***  we shrink, so that we can rely on addresses in the allocation but beyond
***  the end of the array containing zeros.  
**/


int gsetlength (struct amgr *a, int newlen)
{
    char *p;
    int newalloc;

    if (newlen > a->len) {
	if (newlen > a->alloc) {
	    newalloc = newlen * 4 / 3;
	    p = (char *) *a->loc + a->eltsize;
	    p = realloc(p, newalloc * a->eltsize);
	    if (p == NULL) error("gsetlength: realloc failed");
	    bzero(p + a->alloc * a->eltsize, (newalloc - a->alloc) * a->eltsize);
	    a->alloc = newalloc;
	    *a->loc = p - a->eltsize;
	}
	a->len = newlen;
    }
    else if (newlen < a->len) {
	p = (char *) *a->loc + a->eltsize;
	bzero(p + newlen * a->eltsize,
	      (a->len - newlen) * a->eltsize);
	a->len = newlen;
    }
    return 1;
}



/**-------------------------------------------------------------------
***  Create, Read, Write, Free
**/


struct amgr *make_amgr (void **loc, int alloc, int eltsize)
{
    struct amgr *a;
    char *p;

    if (eltsize <= 0) error("make_amgr: Bad eltsize %d", eltsize);
    allocate(a, 1, struct amgr);
    a->len = 0;
    a->alloc = alloc;
    a->eltsize = eltsize;
    a->loc = loc;
    allocate(p, alloc * eltsize, char);
    *a->loc = (void *) (p - eltsize);
    return a;
}


int free_amgr (struct amgr *a)
{
    char *p;

    p = (char *) *a->loc + a->eltsize;
    free(p);
    *a->loc = NULL;
    free(a);
    return 1;
}


#define AmgrCookie 20957881

int write_amgr (struct amgr *a, FILE *file)
{
    char *p;

    write_int(AmgrCookie, file);
    write_int(a->len, file);
    write_int(a->alloc, file);
    write_int(a->eltsize, file);
    p = (char *) *a->loc + a->eltsize;
    write_array(p, a->len * a->eltsize, char, file);
    return 1;
}


struct amgr *read_amgr (void **loc, FILE *file)
{
    struct amgr *a;
    int x, size;
    char *p;

    if (read_int(x, file) == EOF) return NULL;
    if (x != AmgrCookie) error("read_amgr: Not an amgr");
    allocate(a, 1, struct amgr);
    read_int(a->len, file);
    read_int(a->alloc, file);
    read_int(a->eltsize, file);
    if (a->alloc < a->len) error("read_amgr: alloc less than length");
    allocate(p, a->alloc * a->eltsize, char);
    read_array(p, a->len * a->eltsize, char, file);
    a->loc = loc;
    *loc = (void *) (p - a->eltsize);
    return a;
}


void dump_amgr (struct amgr *a, int (*print_elt)(), FILE *file)
{
    int i;
    char *p;
    
    fprintf(file, "Amgr:\n");
    fprintf(file, "len =       %d\n", a->len);
    fprintf(file, "alloc =     %d\n", a->alloc);
    fprintf(file, "eltsize =   %d\n", a->eltsize);
    if (print_elt != NULL) {
	p = (char *) *a->loc + a->eltsize;
	for (i = 0; i < a->len; i++) {
	    fprintf(file, "%-4d: ", i + 1);
	    print_elt(p + i * a->eltsize, file);
	    terpri(file);
	}
    }
}

