#include <stdio.h>
#include "spa.h"


struct dict_node {
    int bit;
    int left;
    int right;
    char *key;
};


struct dict {
    struct dict_node *nodes;
    int count;
    int alloc;
};


#define DICT_STATE_STACKSIZE 128

struct dict_state {
  int stack[DICT_STATE_STACKSIZE];
  char term[DICT_STATE_STACKSIZE];
  int depth;
  struct dict *dict;
};


#define pat_key(i,d) ((d)->nodes[i].key)


int make_node (char *key, int bit);
int pat_search (char *key, struct dict *d);
int pat_search_1 (char *key);
int pat_insert (char *key, struct dict *d);
struct dict *scan_dict (FILE *file);
int pat_encode (struct dict *d, FILE *infile, FILE *outfile);
int pat_encode_fields (int *flds, int len, struct dict *d, FILE *infile, FILE *outfile);
int pat_decode (struct dict *d, FILE *infile, FILE *outfile);
int pat_decode_fields (int *flds, int len, struct dict *d, FILE *infile, FILE *outfile);
struct dict *make_dict (int alloc);
void free_dict (struct dict *d);
void write_dict (struct dict *d, FILE *file);
struct dict *read_dict (FILE *file);
int dict_size (char *filename);
void dump_dict (struct dict *d, FILE *file);
void print_dict_keys (struct dict *d, FILE *file);
void save_dict (struct dict *d, char *filename);
struct dict *load_dict (char *filename);
int dict_start (struct dict_state *state, struct dict *d);
int dict_next (int *kp, char **sp, struct dict_state *S);
int subdict_start (struct dict_state *state, int p, struct dict *d);


struct dict *Dict = NULL;

#define DictType 4345

#define Nodes (Dict->nodes)
#define NodeCount (Dict->count)
#define NodeAlloc (Dict->alloc)
#define NodeInitAlloc (Dict->initalloc)

#define string_bit(b, s, len) (((b)/8 >= len) ? 0 : (s[(b)/8] & 1 << 7-(b)%8))


int make_node (char *key, int bit)
{
    int x = NodeCount++;

    if (x == NodeAlloc) {
	int newalloc = NodeAlloc * 4 / 3;
	reallocate(Nodes, NodeAlloc, newalloc, struct dict_node);
    }
    Nodes[x].key = copy_string(key);
    Nodes[x].bit = bit;
    return x;
}


/**
***  The functions that set Dict have to remember to restore it to its
***  previous value!  Otherwise toes get stepped on in recursive calls.
***  (For example if a function passed to map_dict calls pat_search on
***  a different dict than the first one.)
**/

int pat_search (char *key, struct dict *d)
{
    int x;
    struct dict *prevDict = Dict;

    Dict = d;
    if (NodeCount == 0)
      x = -1;
    else {
      x = pat_search_1(key);
      if (!string_equal(Nodes[x].key, key))
	x = -1;
    }
    Dict = prevDict;
    return x;
}

/**
***  For internal use.  Doesn't fuss with Dict.
**/

int pat_search_1 (char *key)
{
    int x = 0, p, len = strlen(key);
    
    do {
	p = x;
	x = string_bit(Nodes[p].bit, key, len) ? Nodes[p].right : Nodes[p].left;
    }
    while (Nodes[x].bit > Nodes[p].bit);
    return x;
}


int pat_insert (char *key, struct dict *d)
{
    int x = 0, p, y, i, len, len2;
    char *oldkey;
    struct dict *prevDict = Dict;

    Dict = d;
    new_entry_p = 0;

    if (NodeCount == 0) {
      new_entry_p = 1;
      y = make_node(key, 0);
      Nodes[y].left = y;
      Nodes[y].right = y;
    }

    else {
      y = pat_search_1(key);
      oldkey = Nodes[y].key;
      if (!string_equal(key, oldkey)) {

	new_entry_p = 1;
	len = strlen(key); len2 = strlen(oldkey);
	for (i = 0;
	     string_bit(i, key, len) == string_bit(i, oldkey, len2);
	     i++);
	do {
	  p = x;
	  x = string_bit(Nodes[p].bit, key, len) ? Nodes[p].right : Nodes[p].left;
	}
	while (Nodes[x].bit > Nodes[p].bit && Nodes[x].bit < i);
	y = make_node(key, i);
	if (string_bit(i, key, len)) { Nodes[y].right = y; Nodes[y].left = x; }
	else { Nodes[y].left = y; Nodes[y].right = x; }
	if (string_bit(Nodes[p].bit, key, len)) Nodes[p].right = y;
	else Nodes[p].left = y;
      }
    }

    Dict = prevDict;
    return y;
}


struct dict *scan_dict (FILE *infile)
{
  struct dict *d;
  char line[1024];

  d = make_dict(1000);
  while (scan_line(line, 1024, infile) != EOF)
    pat_insert(line, d);
  return d;
}


int pat_encode (struct dict *d, FILE *infile, FILE *outfile)
{
    char line[1024];

    while (scan_line(line, 1024, infile) != EOF) {
      fprintf(outfile, "%d\n", pat_insert(line, d));
    }
}


int pat_encode_fields (int *flds, int len, struct dict *d, FILE *infile, FILE *outfile)
{
    struct record r;
    int k, id;
    
    while (scan_record(&r, infile) != EOF) {
	for (k = 0; k < r.nfields; k++) {
	    if (k != 0) write_char('\t', outfile);
	    if (imember(k+1, flds, len)) {
		id = pat_insert(r.fields[k], d);
		fprintf(outfile, "%d", id);
	    }
	    else {
		fprintf(outfile, "%s", r.fields[k]);
	    }
	}
	write_char('\n', outfile);
    }
}


int pat_decode (struct dict *d, FILE *infile, FILE *outfile)
{
    int x;
    struct dict *prevDict = Dict;

    Dict = d;
    while (scan_int(x, infile) != EOF) {
	fprintf(outfile, "%s\n", Nodes[x].key);
    }
    Dict = prevDict;
}


int pat_decode_fields (int *flds, int len, struct dict *d, FILE *infile, FILE *outfile)
{
    struct record r;
    int k, id;
    struct dict *prevDict = Dict;
    
    Dict = d;
    while (scan_record(&r, infile) != EOF) {
	for (k = 0; k < r.nfields; k++) {
	    if (k != 0) write_char('\t', outfile);
	    if (imember(k+1, flds, len)) {
		id = atoi(r.fields[k]);
		fprintf(outfile, "%s", Nodes[id].key);
	    }
	    else {
		fprintf(outfile, "%s", r.fields[k]);
	    }
	}
	write_char('\n', outfile);
    }
    Dict = prevDict;
}


struct dict *make_dict (int alloc)
{
    struct dict *d;

    allocate(d, 1, struct dict);

    d->count = 0;
    d->alloc = alloc;
    allocate(d->nodes, alloc, struct dict_node);
    return d;
}


void free_dict (struct dict *d)
{
    free(d->nodes);
    free(d);
}


void write_dict (struct dict *d, FILE *file)
{
    int x;
    struct dict_node *n;
    
    write_int(DictType, file);
    write_int(d->count, file);
    write_int(d->alloc, file);

    for (x = 0; x < d->count; x++) {
	n = d->nodes + x;
	write_int(n->bit, file);
	write_int(n->left, file);
	write_int(n->right, file);
	write_string(n->key, file);
    }
}


struct dict *read_dict (FILE *file)
{
    int x;
    struct dict *d;
    struct dict_node *n;

    read_int(x, file);
    if (x != DictType)
      error("read_dict: Not a dict");
    allocate(d, 1, struct dict);
    read_int(d->count, file);
    read_int(d->alloc, file);
    allocate(d->nodes, d->alloc, struct dict_node);

    for (x = 0; x < d->count; x++) {
	n = d->nodes + x;
	read_int(n->bit, file);
	read_int(n->left, file);
	read_int(n->right, file);
	n->key = read_string_alloc(file);
    }
    return d;
}


int dict_size (char *filename)
{
    FILE *file = Fopen(filename, "r");
    int x;

    read_int(x, file);
    if (x != DictType) error("Not a dict: %s", filename);
    read_int(x, file);
    fclose(file);
    return x;
}    


void dump_dict (struct dict *d, FILE *file)
{
    int x;
    struct dict_node *n;
    
    fprintf(file, "Dict *%x\n", d);
    fprintf(file, "count   %d\n", d->count);
    fprintf(file, "alloc   %d\n", d->alloc);
    fprintf(file, "nodes   *%x\n", d->nodes);
    fprintf(file, "  %6s  %8s  %8s  %8s  %s\n", "id", "bit", "left", "right", "key");

    for (x = 0; x < d->count; x++) {
	n = d->nodes + x;
	fprintf(file, "  %6d  %8d  %8d  %8d  \"%s\"\n", x, n->bit, n->left, n->right, n->key);
    }
}


void print_dict_keys (struct dict *d, FILE *file)
{
    int x;
    struct dict_node *n;

    for (x = 0; x < d->count; x++) {
	n = d->nodes + x;
	fprintf(file, "%s\n", n->key);
    }
}


void save_dict (struct dict *d, char *filename)
{
    FILE *file = Fopen(filename, "w");
    write_dict(d, file);
    fclose(file);
}


struct dict *load_dict (char *filename)
{
    FILE *file = Fopen(filename, "r");
    struct dict *d = read_dict(file);
    fclose(file);
    return d;
}


/**
***  Walk over node tree.  This guarantees that keys are visited in
***  sorted order.
**/


int pat_find_parent (char *key)
{
  int p, x = 0, len = strlen(key);

  do {
    p = x;
    x = string_bit(Nodes[p].bit, key, len) ? Nodes[p].right : Nodes[p].left;
  }
  while (Nodes[x].bit > Nodes[p].bit);
  return p;
}


/**  P is taken as a nonterminal node  **/
subdict_start (struct dict_state *state, int p, struct dict *d)
{
  state->depth = 1;
  state->stack[0] = p;
  state->term[0] = 0;
  state->dict = d;
}


dict_start (struct dict_state *state, struct dict *d)
{
  state->dict = d;
  if (d->count == 0) {
    state->depth = 0;
  }
  else {
    state->stack[0] = 0;
    state->term[0] = 0;
    state->depth = 1;
  }
}

int dict_next (int *kp, char **sp, struct dict_state *S)
{
  struct dict_node *nodes = S->dict->nodes;
  int *stack = S->stack, depth = S->depth, p, n;
  char *term = S->term;
  
  if (depth <= 0) return 0;

  while (depth > 0) {
    depth--;
    p = stack[depth];
    if (term[depth]) {
      if (p == 0) {
	if (depth <= 0) error("dict_next: top node is terminal node");
	if (pat_find_parent(nodes[p].key) != stack[depth-1]) continue;
      }	
      S->depth = depth;
      *kp = p;
      *sp = nodes[p].key;
      return 1;
    }
    else {
      n = stack[depth] = nodes[p].right;
      term[depth] = (nodes[n].bit <= nodes[p].bit ? 'a' : 0);
      depth++;
      if (depth >= DICT_STATE_STACKSIZE) error("dict_next: stack overflow");
      n = stack[depth] = nodes[p].left;
      term[depth] = (nodes[n].bit <= nodes[p].bit ? 'a' : 0);
      depth++;
    }
  }
  return 0;
}
