#include <stdio.h>
#include <string.h>
#include "spa.h"


/**
***  These are hash tables for fixed-sized objects, not for strings.
***  Keys and values are both fixed-sized.  Keys are compared using bytecmp.
**/

struct hash_table {
    int size;
    int length;
    int keysize;
    int valsize;
    char *keys;
    char *present;
    char *values;
};

int hash_n_collisions;
int hash_default_size;
#define REHASH_THRESHOLD 0.9
int HashVerbose = 0;


/**
***  choose_good_prime ()
**/

#define N_GOOD_PRIMES 171

int good_primes[N_GOOD_PRIMES] = {
  11, 101, 211, 307, 401, 503, 601, 701, 809, 907, 1009,
  1103, 1201, 1301, 1409, 1511, 1601, 1699, 1787, 1867, 1913,
  1987, 4409, 6997, 9733, 12553, 15401, 18313, 21383, 24281, 27449,
  30559, 33613, 36781, 39971, 43051, 46447, 49663, 52919, 56179, 59359,
  62773, 66103, 69493, 72823, 76207, 79621, 82891, 86371, 89849, 93179,
  96581, 100109, 103591, 107033, 110597, 114113, 117671, 121123, 124577, 128189,
  131701, 135277, 138641, 142169, 145931, 149419, 152879, 156593, 160169, 163841,
  167471, 171169, 174773, 178487, 182059, 185641, 189251, 192847, 196549, 200183,
  203969, 207563, 211093, 214811, 218681, 222289, 225949, 229601, 233341, 237203,
  240967, 244597, 248257, 251947, 255709, 259537, 263401, 267131, 270659, 274529,
  278261, 281947, 285709, 289717, 293507, 297403, 301237, 304901, 308713, 312583,
  316301, 319901, 323707, 327443, 331213, 335033, 338761, 342691, 346469, 350377,
  354073, 357977, 361961, 365689, 369709, 373661, 377459, 381373, 385199, 389171,
  392849, 396637, 400559, 404539, 408607, 412303, 416281, 420163, 424003, 427991,
  431803, 435427, 439519, 443231, 447217, 451039, 455233, 459113, 463319, 467213,
  470887, 474769, 478631, 482507, 486653, 490459, 494383, 498257, 502087, 506131,
  510079, 514061, 517919, 521789, 525727, 529681, 533693, 537661, 541699, 545747 };


int choose_good_prime (hint)
    int hint;
{
    int i;

    /* A binary search would be faster */
    for (i = 0; ; i++) {
	if (i >= N_GOOD_PRIMES) error("Out of good primes");
	if (good_primes[i] >= hint) {
	    return good_primes[i];
	}
    }
}

/**
***  keys[-1] and values[-1] are temporary holding cells
**/

struct hash_table *make_hash_table (int size, int keysize, int valsize)
{
    struct hash_table *table;
    char *keys, *present, *values;

    allocate(table, 1, struct hash_table);
    size = choose_good_prime(size);
    table->size = size;
    table->length = 0;
    table->keysize = keysize;
    table->valsize = valsize;
    allocate(keys, keysize * (size + 2), char);
    allocate(table->present, (size - 1) / 8 + 1, char);
    allocate(values, valsize * (size + 2), char);
    table->keys = keys + 2 * keysize;
    table->values = values + 2 * valsize;
    return table;
};


int free_hash_table (struct hash_table *table)
{
    free(table->keys - 2 * table->keysize);
    free(table->values - 2 * table->valsize);
    free(table->present);
    free(table);
}


void rehash (struct hash_table *table)
{
    error("Haven't implemented rehashing");
}


void clrhash (struct hash_table *table)
{
    bzero(table->present, (table->size - 1) / 8 + 1);
    table->length = 0;
}


#define KEYBITSET(table, i) (table->present[i/8] & (1 << (i%8)))
#define SETKEYBIT(table, i) (table->present[i/8] |= (1 << (i%8)))
#define UNSETKEYBIT(table, i) (table->present[i/8] &= ~(1 << (i%8)))


void print_bytes (char *b, int n, FILE *file)
{
    int i, start = 1;

    while (n > 0) {
	if (start) start = 0; else fprintf(file, " ");
	fprintf(file, "%02x", *b);
	b++; n--;
    }
}
    

void print_hash_table_header (struct hash_table *table, FILE *file)
{
    fprintf(file, "Hash Table *%d\n", table);
    fprintf(file, "size    = %d\n", table->size);
    fprintf(file, "length  = %d\n", table->length);
    fprintf(file, "keysize = %d\n", table->keysize);
    fprintf(file, "valsize = %d\n", table->valsize);
}

void dump_hash_table (struct hash_table *table, FILE *file)
{
    char *keys = table->keys, *values = table->values;
    int i, keysize = table->keysize, valsize = table->valsize;

    fprintf(file, "P: ");
    print_bytes(table->present, (table->size - 1) / 8 + 1, file);
    fprintf(file, "\n");
    for (i = 0; i < table->size; i++) {
	if (KEYBITSET(table, i)) {
	    fprintf(file, "%4d: ", i);
	    print_bytes(keys + i * keysize, keysize, file);
	    fprintf(file, " -> ");
	    print_bytes(values + i * valsize, valsize, file);
	    fprintf(file, "\n");
	}
    }
}


/**
***  This lays down the bits from the key onto a long in sweeps forward and
***  backward, to counteract the on bits that bunch at the low end.
**/

void hash_keys (char *b, int n, int tblsize, int *h1, int *h2)
{
    int i = 0, off = 0;
    unsigned long k = 0L;
    char *out = (char *)&k;

    while (n > 0) {
	out[(i+off) % sizeof(long)] ^= *b++;
	n--;
	i++;
	if (i == sizeof(long)) {
	    i = 0;
	    off++;
	}
    }

    *h1 = k % tblsize;
    *h2 = (k / tblsize) % tblsize + 1;
    if (*h2 == tblsize) *h2 = 1;

#ifdef DEBUG
    if (HashVerbose) printf("k= %d,  tblsize= %d,  h1= %d,  h2= %d\n", k, tblsize, *h1, *h2);
#endif
}


/**
***  We use double hashing to resolve conflicts.  H1() gives us the first
***  index to probe.  H2() gives us the decrement for the probe path.
*** 
***  We know that:
***  
***    1. There is an empty cell in the table
***    2. Objects along a probe sequence are in decreasing order.
***  
***  Thus, if we encounter a NULL key or a key that is less than the search
***  key, the search key isn't in the table.
***   
***  hash_index_1(key, start, table) returns the index at which the search key
***  belongs if we start looking at position <start>.
***  
***  We need to define key_to_int() and key_cmp().  Every key should be 
***  greater than 0.
**/

/**
***  Probe the path starting at <i>, skipping by <decr>, until the search
***  key is less than or equal to the probed key.  This is the position at
***  which the probe key belongs.
**/

/**  Find the place in the table where the key belongs.  If it is already
***  present, hash_key_found is 1, else 0.
**/


#define START -1
#define EMPTY 1
#define FOUND 2
#define DISPLACE 3


int find_loc (char *key, struct hash_table *table, int start, int *status)
{
    char *keys = table->keys;
    int keysize = table->keysize, tblsize = table->size;
    int h1, h2, i, cmp;

    hash_keys(key, keysize, tblsize, &h1, &h2);

    if (start == START) {
	i = h1;
    }
    else {
	i = start - h2;
	if (i < 0) i += tblsize;
    }

#ifdef DEBUG
    if (HashVerbose) printf("Probe Path: %d -%d\n", i, h2);
#endif
    while (1) {
#ifdef DEBUG
	if (HashVerbose) printf("  Probe %d\n", i);
#endif
	if (!KEYBITSET(table, i)) {
#ifdef DEBUG
	    if (HashVerbose) printf("    Found empty cell\n", i);
#endif
	    *status = EMPTY;
	    return i;
	}
#ifdef DEBUG
	if (HashVerbose) {
	    printf("    Collision:  new= ", i);
	    print_bytes(key, keysize, stdout);
	    printf("  old= ");
	    print_bytes(keys + i * keysize, keysize, stdout);
	    printf("\n");
	}
#endif
	hash_n_collisions++;
	cmp = bcmp(key, keys + i * keysize, keysize);
	if (cmp < 0) {
#ifdef DEBUG
	    if (HashVerbose) printf("    New key smaller, displace old\n");
#endif
	    *status = DISPLACE;
	    return i;
	}
	else if (cmp == 0) {
#ifdef DEBUG
	    if (HashVerbose) printf("    Keys equal: key found\n");
#endif
	    *status = FOUND;
	    return i;
	}
	else {
#ifdef DEBUG
	    if (HashVerbose) printf("    New key larger, continuing search\n");
#endif
	    i -= h2;
	    if (i < 0) i += tblsize;
	}
    }
}


/**  Displace the key at position i.  If the new position is occupied, displace
***  that key.  Keep going till you find an empty position.
**/

int displace_key (int i, struct hash_table *table)
{
    int keysize = table->keysize, valsize = table->valsize, j, status;
    char *keys = table->keys, *tmpkey1 = keys - keysize, *tmpkey2 = tmpkey1 - keysize;
    char *values = table->values, *tmpvalue1 = values - valsize, *tmpvalue2 = tmpvalue1 - valsize;

    bcopy(keys + i * keysize, tmpkey1, keysize);
    bcopy(values + i * valsize, tmpvalue1, valsize);
    UNSETKEYBIT(table, i);

#ifdef DEBUG
    if (HashVerbose) fprintf(stderr, "Displace key from %d\n", i);
#endif
    while (1) {
	j = find_loc(tmpkey1, table, i, &status);
	switch (status) {
	  case FOUND:
	    error("Displaced key found in another place in table.");
	  case EMPTY:
#ifdef DEBUG
	    if (HashVerbose) printf("  Insert displaced key at %d\n", i);
#endif
	    bcopy(tmpkey1, keys + j * keysize, keysize);
	    bcopy(tmpvalue1, values + j * valsize, valsize);
	    SETKEYBIT(table, j);
	    return 1;
	  case DISPLACE:
#ifdef DEBUG
	    if (HashVerbose) printf("  Insert displaced key at %d, displace that one\n", i);
#endif
	    bcopy(keys + j * keysize, tmpkey2, keysize);
	    bcopy(values + j * valsize, tmpvalue2, valsize);
	    bcopy(tmpkey1, keys + j * keysize, keysize);
            bcopy(tmpvalue1, values + j * valsize, valsize);
	    bcopy(tmpkey2, tmpkey1, keysize);
	    bcopy(tmpvalue2, tmpvalue1, valsize);
	    break;
	  default:
	    error("Bad status");
	}
    }
}


/**
***  Find the key in the table.
**/

void *gethash (void *key, struct hash_table *table)
{
    int status, i;

    i = find_loc(key, table, START, &status);
    return (status == FOUND ? table->values + i * table->valsize : NULL);
}


/**
***  Find the key in the table.  If not found, insert the key at the appropriate
***  index, displacing the key there if necessary.  Return a pointer to
***  the value.  (The value is 0 if the key is newly inserted).
**/

extern int new_entry_p;

void *needhash (void *key, struct hash_table *table)
{
    int status, i;

    new_entry_p = 0;
    i = find_loc(key, table, START, &status);
    if (status == FOUND)
      return table->values + i * table->valsize;

    new_entry_p = 1;
    /* Filled entries + new entry + one empty entry */
    if (table->length + 2 > table->size * REHASH_THRESHOLD) {
	rehash(table);
	i = find_loc(key, table, START, &status);
    }
    if (status == DISPLACE)
      displace_key(i, table);
    else if (status != EMPTY)
      error("Bad status");
    bcopy(key, table->keys + i * table->keysize, table->keysize);
    SETKEYBIT(table, i);
    table->length++;
    return table->values + i * table->valsize;
}


/**
***  Set the value of key in table.  Return value.
**/

void *sethash (void *key, void *value, struct hash_table *table)
{
    void *vloc = needhash(key, table);
    bcopy(value, vloc, table->valsize);
    return vloc;
}


/** f is called on each (key,value,i) triple in apparently random order.
**/

void maphash (int (*f)(), struct hash_table *table)
{
    int i, size = table->size;
    char *keys = table->keys, *values = table->values;
    char *key, *value;

    for (i = 0; i < size; i++) {
	if (KEYBITSET(table, i)) {
	    key = table->keys + i * table->keysize;
	    value = table->values + i * table->valsize;
	    f(key, value, i);
	}
    }
}


/*
 * main ()
 * {
 *     int i, k, v, *vp, ents[5][2] = {
 * 	{ 6, 20 },
 * 	{ 10, -32 },
 * 	{ 0, 8 },
 * 	{ 10, 0 },
 * 	{ 17, 1 }
 *     };
 *     struct hash_table *table = make_hash_table(10, sizeof(int), sizeof(int));
 * 
 *     for (i = 0; i < 5; i++) {
 * 	k = ents[i][0];
 * 	v = ents[i][1];
 * 	printf("%d %d\n", k, v);
 * 	sethash(&k, &v, table);
 * 	dump_hash_table(table, stdout);
 * 	printf("\n");
 * 
 * 	printf("gethash %d\n", k);
 * 	vp = gethash(&k, table);
 * 	if (vp == NULL)	printf("Return = NULL\n", k);
 * 	else printf("Return = %d\n", *vp);
 * 	printf("\n");
 *     }
 * }
 */


/**
***  struct hash_table {
***      int size;
***      int length;
***      int keysize;
***      int valsize;
***      char *keys;
***      char *present;
***      char *values;
***  };
**/

write_hash_table (struct hash_table *t, FILE *file)
{
    write_int(t->size, file);
    write_int(t->length, file);
    write_int(t->keysize, file);
    write_int(t->valsize, file);
    write_bytes(t->keys, t->keysize * t->size, file);
    write_bytes(t->present, (t->size-1)/8+1, file);
    write_bytes(t->values, t->valsize * t->size, file);
}

struct hash_table *read_hash_table (FILE *file)
{
    struct hash_table *t;

    allocate(t, 1, struct hash_table);
    read_int(t->size, file);
    read_int(t->length, file);
    read_int(t->keysize, file);
    read_int(t->valsize, file);
    allocate(t->keys, t->keysize * (t->size + 2), char);
    allocate(t->present, (t->size-1)/8+1, char);
    allocate(t->values, t->valsize * (t->size + 2), char);
    t->keys += t->keysize * 2;
    t->values += t->valsize * 2;
    read_bytes(t->keys, t->keysize * t->size, file);
    read_bytes(t->present, (t->size-1)/8+1, file);
    read_bytes(t->values, t->valsize * t->size, file);
    return t;
}

