#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <error.h>
#include <spastrng.h>   /* string_equal */


#define MAXNFLAGS 128
#define MAXNARGS 32
char *Usage = NULL;
char *ProgName = NULL;


struct pa_data {
    char *ofmt;                    /* the original format string */
    char *fmt;                     /* after pa_get_locs, just the nonflag part */
    char ***argvp;
    int *argcp;
    char *flags[MAXNFLAGS];        /* flag parameters */
    int nflags;                    /* number of flag parameters */
    void *flaglocs[MAXNFLAGS];     /* locations to receive flag arguments */
    void *arglocs[MAXNARGS];       /* locations to receive nonflag arguments */
    int nargs;                     /* number of nonflag parameters */
    int moreallowed;
};    

#define OFmt        (PaData->ofmt)
#define Fmt         (PaData->fmt)
#define ArgcP       (PaData->argcp)
#define ArgvP       (PaData->argvp)
#define Argc        (*PaData->argcp)
#define Argv        (*PaData->argvp)
#define Flags       (PaData->flags)
#define FlagLocs    (PaData->flaglocs)
#define NFlags      (PaData->nflags)
#define ArgLocs     (PaData->arglocs)
#define NArgs       (PaData->nargs)
#define MoreAllowed (PaData->moreallowed)

#define HANDLED     0
#define NOT_HANDLED 1


pa_get_locs (struct pa_data *PaData, va_list args)
{
    char *s;

    NArgs = NFlags = MoreAllowed = 0;

    /* Get arg locs */
    for (s = Fmt; *s && *s != '-' && *s != '*'; s++) {
	if (*s != '?') {
	    if (NArgs >= MAXNARGS)
	      error("parse_args: arg buffer overflow");
	    ArgLocs[NArgs++] = va_arg(args, void *);
	}
    }
    if (*s == '*') {
	*s++ = '\0';
	MoreAllowed = 1;
	if (*s && *s != '-') error("parse_args: bad fmt string \"%s\" (after '*')", OFmt);
    }
      
    /* Get flag locs */
    while (*s) {
	*s++ = '\0';
	if (NFlags >= MAXNFLAGS) error("parse_args: flag buffer overflow");
	if (*s == '\0' || *(s+1) == '\0')
	  error("parse_args: bad fmt string \"%s\" (missing flag type or name)", OFmt);
	Flags[NFlags] = s;
	FlagLocs[NFlags++] = va_arg(args, void *);
	while (*s && *s != '-') s++;
    }
}


pa_bind_flag (int k, struct pa_data *PaData)
{
    char *arg;
    void *loc;

    loc = FlagLocs[k];
    if (Flags[k][0] == 'b') *(int *) loc = 1;
    else {
	if (Argc == 0) {
	    fprintf(stderr, "%s: Missing arg for flag %s\n", ProgName, Flags[k]+1);
	    if (Usage != NULL) fprintf(stderr, "Usage: %s\n", Usage);
	    exit(1);
	}
	arg = *Argv++; Argc--;
	switch (Flags[k][0]) {
	  case 'c': *(char *) loc = *arg; break;
	  case 'i': *(int *) loc = atoi(arg); break;
	  case 'l': *(long *) loc = atol(arg); break;
	  case 'f': *(float *) loc = atof(arg); break;
	  case 'd': *(double *) loc = atof(arg); break;
	  case 'D': *(long double *) loc = atof(arg); break;
	  case 's': *(char **) loc = arg; break;
	  default: error("parse_args: Unrecognized type spec '%c' in \"%s\"",
			 Flags[k][0], OFmt);
	}
    }
}


pa_handle_flag (char *flag, struct pa_data *PaData)
{
    int i, k;

    for (k = 0; k < NFlags; k++) {

	/* known flag */
	if (string_equal(flag, Flags[k]+1)) {
	    Argc--, Argv++;
	    pa_bind_flag(k, PaData);
	    return HANDLED;
	}
    }

    /* unknown flag */
    /* debug */
    if (string_equal(flag, "debug")) {
	Argc--, Argv++;
	debug_on_error = 1;
	return HANDLED;
    }

    /* the nonflag handler won't be able to do anything with this,           */
    /* so we're allowed to quit only if there are no obligatory nonflag args */
    if (MoreAllowed == 1 && (Fmt[0] == '\0' || Fmt[0] == '?')) return NOT_HANDLED;

    /* concatenation of single-character flags */
    Argc--, Argv++;
    for (i = 0; flag[i]; i++) {
	for (k = 0; k < NFlags; k++) {
	    if (Flags[k][1] == flag[i] && Flags[k][2] == 0) {
		pa_bind_flag(k, PaData);
		break;
	    }
	}
	if (k == NFlags) {
	    fprintf(stderr, "%s: unrecognized flag at '%c' in \"-%s\"\n", ProgName, flag[i], flag);
	    if (Usage != NULL) fprintf(stderr, "Usage: %s\n", Usage);
	    exit(1);
	}
    }
    return HANDLED;
}


char *parse_args (int *acp, ...)
{
    va_list args;
    struct pa_data databuf, *PaData = &databuf;
    int doing_optionals = 0, n;
    char *arg;
    void *loc;

    va_start(args,acp);
    ArgcP = acp;
    ArgvP = va_arg(args, char ***);
    OFmt = va_arg(args, char *);
    /* strdup because this may be a constant string, but we'll want to modify it */
    Fmt = strdup(OFmt);
    ProgName = *Argv++; Argc--;

    /* usage request */
    if (Argc == 1 && (string_equal(Argv[0], "-usage") || string_equal(Argv[0], "-?"))) {
	printf("%s\n", Usage == NULL ? OFmt : Usage);
	exit(0);
    }

    pa_get_locs(PaData,args);
    va_end(args);

    /* Parse */
    n = 0;			/* n of nonflag args seen */
    while (Argc > 0) {

	/* flag arg */
	if (Argv[0][0] == '-' && (Argv[0][1] != '\0' && Argv[0][1] != '-')) {
	    if (pa_handle_flag(Argv[0]+1, PaData) == NOT_HANDLED) goto TERMINATE;
	}

	/* nonflag arg */
	else {
	    if (*Fmt == '?') Fmt++, doing_optionals = 1;
	    if (*Fmt == '\0') {
		if (MoreAllowed) goto TERMINATE;
		fprintf(stderr, "%s: Too many args supplied\n", ProgName);
		if (Usage != NULL) fprintf(stderr, "Usage: %s\n", Usage);
		exit(1);
	    }
	    Argc--;
	    arg = *Argv++;
	    if (arg[0] == '-') arg++;
	    loc = ArgLocs[n++];
	    switch (*Fmt++) {
	      case 'c': *(char *) loc = *arg; break;
	      case 'i': *(int *) loc = atoi(arg); break;
	      case 'l': *(long *) loc = atol(arg); break;
	      case 'f': *(float *) loc = atof(arg); break;
	      case 'd': *(double *) loc = atof(arg); break;
	      case 'D': *(long double *) loc = atof(arg); break;
	      case 's': *(char **) loc = arg; break;
	      case 'b': error("parse_args: Arg type 'b' only applicable to flags");
	      default: error("parse_args: Unrecognized type spec '%c'", *Fmt);
	    }
	}
    }

  TERMINATE:
    if (*Fmt != '\0' && *Fmt != '?' && !doing_optionals) {
	fprintf(stderr, "%s: Not enough args supplied\n", ProgName);
	if (Usage != NULL) fprintf(stderr, "Usage: %s\n", Usage);
	exit(1);
    }

    return ProgName;
}
