#include <stdio.h>
#include "spa.h"
#include "word.h"

/**************************************************************************/
/**                                                                      **/
/**  A pool consists of a linked list of poolsegs.  The pool struct      **/
/**  contains a pointer to the first and last poolseg, and the           **/
/**  allocation for new poolsegs (in chars).  When we can't satisfy an   **/
/**  allocation request in a poolseg, we go to the next one.  When       **/
/**  there is none left, we allocate a new one.  The size of the new     **/
/**  one is pool->alloc + sizeof(struct poolseg).                        **/
/**                                                                      **/
/**  Since what malloc returns is correctly aligned for everything, we   **/
/**  don't need to worry about alignment for the initial allocation.     **/
/**  Thereafter, we alloc things in units of words, to avoid alignment   **/
/**  problems.  That is, allocations are rounded up to the nearest       **/
/**  multiple of the word size.  My assumption is that that will         **/
/**  satisfy alignment requirements.                                     **/
/**                                                                      **/
/**  The len of a poolseg is the number of chars that have been          **/
/**  allocated in the allocation area (excluding the header).  It may    **/
/**  not exceed pool->alloc.                                             **/
/**                                                                      **/
/**************************************************************************/


struct pool { int alloc; struct poolseg *first; struct poolseg
*last; };

struct poolseg {
    int len;
    struct poolseg *next;
};

static struct poolseg *make_poolseg (struct pool *pool)
{
    struct poolseg *seg;

    seg = malloc(pool->alloc + sizeof(struct poolseg));
    seg->len = 0;
    seg->next = NULL;
    return seg;
}

struct pool *make_pool (int alloc)
{
    struct pool *pool;

    allocate(pool, 1, struct pool);
    pool->alloc = alloc;
    pool->last = pool->first = make_poolseg(pool);
    return pool;
}

zero_pool (struct pool *pool)
{
    pool->first->len = 0;
    pool->last = pool->first;
}

#define poolseg_base(seg)   ((char *)((seg)+1))
#define round_alloc(alloc)  ((((alloc)-1)/sizeof(word) + 1) * sizeof(word))

void *palloc (int nelts, int eltsize, struct pool *pool)
{
    struct poolseg *seg = pool->last;
    int newlen, alloc;
    char *b;

    alloc = round_alloc(nelts*eltsize);
    if (alloc > pool->alloc) error("palloc: allocation exceeds maximum pool allocation");

    newlen = seg->len + alloc;
    if (newlen > pool->alloc) {
	if (seg->next == NULL) seg->next = make_poolseg(pool);
	else seg->next->len = 0;
	seg = seg->next;
	pool->last = seg;
	newlen = alloc;
    }
    b = poolseg_base(seg) + seg->len;
    bzero(b, alloc);
    seg->len = newlen;
    return (void *) b;
}
