#include <stdio.h>
#include <limits.h>
#include "spa.h"
#include "aarray.h"

struct tnode {
  int key;
  int left;
  int right;
};

struct tstore {
  struct aarray *nodes;
  struct aarray *colors;
};

struct tstore *TStore = NULL;
struct tstore *TStoreStack[32];
int TStoreStackDepth = 0;
int TMarkDepth = 0;

static struct tnode *TNodes;
static char *TColors;

static int Key, GG, G, P, X;
extern int new_entry_p;

#define TAIL 0
#define HEAD 1

#define BLACK 0
#define RED 1

#define REDP(x)  (TColors[(x)/8] & 1 << (x) % 8)
#define REDDEN(x)   (TColors[(x)/8] |= 1 << (x) % 8)
#define BLACKEN(x) (TColors[(x)/8] &= ~(1 << (x) % 8))

static int make_node (int key)
{
  struct aarray *nodes = TStore->nodes, *colors = TStore->colors;
  int new = nodes->len;
  struct tnode *n = apull(nodes);

  n->key = key;
  if (nodes->len > colors->len * 8) apull(colors);
  return new;
}

struct tstore *make_tstore (int alloc)
{
  struct tstore *store, *prev = TStore;
  struct tnode *nodes;
  char *colors;

  allocate(store, 1, struct tstore);
  store->nodes = make_aarray(alloc, sizeof(struct tnode));
  store->colors = make_aarray(alloc/8, 1);
  TStore = store;
  if (make_node(0) != TAIL) error("make_tstore: This shouldn't happen");
  if (make_node(INT_MIN) != HEAD) error("make_tstore: This shouldn't happen either");
  TStore = prev;
  return store;
}

struct tstore *zero_tstore (struct tstore *tstore)
{
  azero_to(2, tstore->nodes);
  azero_to(1, tstore->colors);
  tstore->colors->contents[0] = 0;
  return tstore;
}

static int rotate (int g)
{
  int x, p = Key < TNodes[g].key ? TNodes[g].left : TNodes[g].right;
  
  if (Key < TNodes[p].key) {
    x = TNodes[p].left;
    TNodes[p].left = TNodes[x].right;
    TNodes[x].right = p;
  }
  else {
    x = TNodes[p].right;
    TNodes[p].right = TNodes[x].left;
    TNodes[x].left = p;
  }
  if (Key < TNodes[g].key) TNodes[g].left = x;
  else TNodes[g].right = x;
  return x;
}

static int split ()
{
  REDDEN(X);
  BLACKEN(TNodes[X].left);
  BLACKEN(TNodes[X].right);
  if (REDP(P)) {
    REDDEN(G);
    if (Key < TNodes[G].key != Key < TNodes[P].key)
      P = rotate(G);
    X = rotate(GG);
    BLACKEN(X);
  }
  BLACKEN(TNodes[HEAD].right);
}

/**  Initially, loc should contain TAIL (the null tree)  **/
int tinsert (int key, int *loc)
{
  int new;

  TNodes = (struct tnode *) TStore->nodes->contents;
  TColors = (char *) TStore->colors->contents;
  Key = key;
  TNodes[HEAD].right = *loc;
  X = HEAD;
  new_entry_p = 0;

  P = X; G = X;
  do {
    GG = G; G = P; P = X;
    if (X != HEAD && Key == TNodes[X].key) {
	*loc = TNodes[HEAD].right;
	return X;
    }
    if (Key < TNodes[X].key) X = TNodes[X].left;
    else X = TNodes[X].right;
    if (REDP(TNodes[X].left) && REDP(TNodes[X].right))
      split();
  }
  while (X != TAIL);
  X = new = make_node(Key);
  new_entry_p = 1;
  /* Because the new node could cause a reallocation */
  TNodes = (struct tnode *) TStore->nodes->contents;
  TColors = (char *) TStore->colors->contents;
  if (Key < TNodes[P].key) TNodes[P].left = X;
  else TNodes[P].right = X;
  split();
  *loc = TNodes[HEAD].right;
  return new;
}

int Tsearch (int key, int t, struct tstore *tstore)
{
    TNodes = (struct tnode *) tstore->nodes->contents;

    while (t != TAIL) {
	if (TNodes[t].key == key) return t;
	else if (key < TNodes[t].key) t = TNodes[t].left;
	else t = TNodes[t].right;
    }
    return 0;
}

dump_tstore (struct tstore *tstore, FILE *file)
{
  int i;
  
  TNodes = (struct tnode *) tstore->nodes->contents;
  TColors = (char *) tstore->colors->contents;
  fprintf(file, "Tstore:\n", tstore);
  fprintf(file, "%d nodes, %d bytes color\n", tstore->nodes->len, tstore->colors->len);
  fprintf(file, "%6s R %6s %6s %6s\n", "node", "left", "key", "right");
  for (i = 0; i < tstore->nodes->len; i++) {
    fprintf(file, "%6d %c %6d %6d %6d\n",
	    i, REDP(i) ? '+' : '-', TNodes[i].left, TNodes[i].key, TNodes[i].right);
  }
}

int scan_tree (FILE *file)
{
  int key, tree = TAIL;

  while (scan_int(key, file) != EOF) {
    tinsert(key, &tree);
  }
  return tree;
}

#define TREE_STACKSIZE 64

struct tree_state {
  int stack[TREE_STACKSIZE];
  char nstate[TREE_STACKSIZE];
  int depth;
};
  
tlist_start (struct tree_state *state, int tree)
{
  if (tree == TAIL)
    state->depth = 0;
  else {
    state->stack[0] = tree;
    state->nstate[0] = 0;
    state->depth = 1;
  }
}

int tlist_next (int *kp, struct tree_state *state)
{
  int *stack = state->stack, depth = state->depth, x, n;
  char *nstate = state->nstate;

  TNodes = (struct tnode *) TStore->nodes->contents;

  while (depth > 0) {
    x = stack[depth-1];
    if (nstate[depth-1] == 0) {
      nstate[depth-1] = 1;
      if ((n = TNodes[x].left) != TAIL) {
	if (depth >= TREE_STACKSIZE) error("tlist_next: stack overflow");
	nstate[depth] = 0;
	stack[depth++] = n;
      }
    }
    else {
      if ((n = TNodes[x].right) != TAIL) {
	stack[depth-1] = n;
	nstate[depth-1] = 0;
      }
      else depth--;
      state->depth = depth;
      *kp = TNodes[x].key;
      return x;
    }
  }
  return 0;
}

int tlist_last (int *kp, int t)
{
  int n;

  TNodes = (struct tnode *) TStore->nodes->contents;

  while ((n = TNodes[t].right) != TAIL) t = n;
  if (kp != NULL) *kp = TNodes[t].key;
  return t;
}

int Tlist_length (int t, struct tstore *tstore)
{
    struct tstore *prev = TStore;
    struct tree_state state;
    int key, len = 0;

    TStore = tstore;
    tlist_start(&state, t);
    while (tlist_next(&key, &state)) len++;
    TStore = prev;
    return len;
}

int print_tlist (int tree, FILE *file)
{
  struct tree_state state;
  int key, first = 1;

  tlist_start(&state, tree);
  fprintf(file, "(");
  while (tlist_next(&key, &state)) {
    if (first) first = 0;
    else fprintf(file, " ");
    fprintf(file, "%d", key);
  }
  fprintf(file, ")");
}

int tintersect (int t1, int t2, int *ctp, int *tp)
{
  struct tree_state s1, s2;
  int k1, k2;

  *tp = NIL;
  *ctp = 0;
  tlist_start(&s1, t1);
  tlist_start(&s2, t2);
  if (tlist_next(&k1, &s1) && tlist_next(&k2, &s2)) {
    while (1) {
      if (k1 == k2) {
	(*ctp)++;
	tinsert(k1, tp);
	if (!(tlist_next(&k1, &s1) && tlist_next(&k2, &s2)))
	  break;
      }
      else if (k1 < k2) {
	if (!tlist_next(&k1, &s1)) break;
      }
      else {
	if (!tlist_next(&k2, &s2)) break;
      }
    }
  }
}

/**  Destructively add all members of t2 to t1.  t2 unchanged. */
int tnunion (int *tp1, int t2)
{
  int elt;
  struct tree_state S;

  tlist_start(&S, t2);
  while (tlist_next(&elt, &S))
    tinsert(elt, tp1);
}

int tnunion1 (int *tp1, struct tstore *tstore1, int t2, struct tstore *tstore2)
{
  int elt;
  struct tree_state S;
  struct tstore *prev = TStore;

  TStore = tstore2, tlist_start(&S, t2);
  while (TStore = tstore2, tlist_next(&elt, &S))
    TStore = tstore1, tinsert(elt, tp1);
  TStore = prev;
}


/*
 * main ()
 * {
 *   int t;
 *   
 *   push_tstore(make_tstore(100));
 *   print_tlist(scan_tree(stdin), stdout);
 *   terpri(stdout);
 *   dump_tstore(TStore, stdout);
 *   terpri(stdout);
 * }
 */




int write_tstore (struct tstore *tstore, FILE *file)
{
    write_aarray(tstore->nodes, file);
    write_aarray(tstore->colors, file);
}

struct tstore *read_tstore (FILE *file)
{
    struct tstore *tstore;

    allocate(tstore, 1, struct tstore);
    tstore->nodes = read_aarray(file);
    tstore->colors = read_aarray(file);
    return tstore;
}
