#include <stdio.h>
#include <limits.h>
#include <ctype.h>

#include <spa.h>
#include <regfsa.h>

char *RegFilename = NULL;
char *FscFilename = NULL;


/**===================================================================
***
***				Level
***
***  This represents a level in the cascade.  The name is a category.
***  Fsr is an fsa (index), fst0 is an fsa, and nfsts tells how many
***  fsts there are for this level.  They are all sequential: if fst0
***  is 4 and nfsts is 3, then the fsts for this level are fsas 4, 5,
***  and 6.
**/

typedef int LevelId;

struct level {
    SymbolId name;   /* A cat */
    FsaId fsr;       /* FSR, merged */
    FsaId fst1;      /* First FST, one for each pattern */
    FsaId fstN;      /* Last FST (does not include the merged FSR) */
    SymbolId out1;
    SymbolId outN;
    ArrayId fstmap;  /* List: cat->fst */
};

struct level *Levels = NULL;
struct amgr *LevelMgr;

#define NLevels  (glength(LevelMgr))

#define level_name(lvl)   (Levels[lvl].name)
#define level_fsr(lvl)    (Levels[lvl].fsr)
#define level_fst1(lvl)   (Levels[lvl].fst1)
#define level_fstN(lvl)   (Levels[lvl].fstN)
#define level_out1(lvl)   (Levels[lvl].out1)
#define level_outN(lvl)   (Levels[lvl].outN)
#define level_fstmap(lvl) (Levels[lvl].fstmap)

#define new_level()       (gnew(LevelMgr))

#define init_levels()     (LevelMgr = make_amgr(&Levels, 20, sizeof(struct level)))

/* we want &fstmap[out1] = Ints + i */
#define make_fstmap(out1,nouts)  (galloc(IntMgr,(nouts)) - (out1))

LevelId CurrentLevel = 0;
SymbolId LevelName, Out1, OutN;
int *FstMap = NULL;


int use_level (LevelId lvl)
{
    CurrentLevel = lvl;
    LevelName = level_name(lvl);
    Out1 = level_out1(lvl);
    OutN = level_outN(lvl);
    FstMap = Ints + level_fstmap(lvl);
}

SymbolId CatFstCat;
#define cat_fst(cat) \
  (CatFstCat = (cat), \
   CatFstCat < Out1 || CatFstCat > OutN ? 0 \
   : FstMap[CatFstCat])

LevelId find_cat_level (SymbolId cat)
{
    LevelId i, prev = CurrentLevel;
    
    for (i = 1; i <= NLevels; i++) {
	use_level(i);
	if (cat_fst(cat)) {
	    if (prev) use_level(prev);
	    return i;
	}
    }
    if (prev) use_level(prev);
    return 0;
}



/**===================================================================
***
***			      SCAN LEVEL
**/


int scan_level (struct tokstream *file)
{
    int tt;
    RegexId rx, rx1, rxN;
    FsaId fsa, fst1, fstN;
    LevelId lvl;
    SymbolId cat;

    init_regices();
    if (Fsas == NULL) init_fsas();
    fst1 = NextFsa;

    if (peek_token(file) != COLON) return 0;
    scan_token(file);    /* the colon */
    tt = scan_token(file);
    if (tt != STRING) {
	error("%s, line %d, tok %s: expecting level name",
	      file->filename, file->line, toktype_string(tt));
    }
    LevelName = intern(file->token);
    rx1 = NextRegex;
    while ((tt = peek_token(file)) != COLON && tt != EOF) {
	rx = scan_regex(file);
	if (!regex_is_macro(rx)) fsa = compile_regex(rx);
    }
    rxN = rx;
    fstN = fsa;

    if (rx1 == NextRegex) {
	error("%s, line %d: No patterns in level %s",
	      file->filename, file->line, symbol_string(LevelName));
    }
    if (Levels == NULL) init_levels();
    lvl = new_level();
    level_name(lvl) = LevelName;
    level_fst1(lvl) = fst1;
    level_fstN(lvl) = fstN;
  
    if (PrintCompileMessages)
      printf("FSR %s\n", symbol_string(LevelName));
    rx = combine_regices(LevelName, rx1, rxN);
    delete_actions(rx);
    level_fsr(lvl) = compile_regex(rx);

    Out1 = INT_MAX;
    OutN = INT_MIN;
    for (fsa = fst1; fsa <= fstN; fsa++) {
	cat = fsa_cat(fsa);
	if (cat == 0) error("scan_level: Fst with null cat");
	if (fsa_nlocals(fsa) > fsa_ilocals(fsa)) {
	    if (cat < Out1) Out1 = cat;
	    if (cat > OutN) OutN = cat;
	}
    }
    level_out1(lvl) = Out1;
    level_outN(lvl) = OutN;
    level_fstmap(lvl) = make_fstmap(Out1, OutN - Out1 + 1);
    FstMap = Ints + level_fstmap(lvl);
    
    for (fsa = fst1; fsa <= fstN; fsa++) {
	if (fsa_nlocals(fsa) > fsa_ilocals(fsa)) {
	    FstMap[fsa_cat(fsa)] = fsa;
	}
    }

    return lvl;
}


dump_levels (FILE *file)
{
    LevelId lvl;
    SymbolId cat;
    int *fstmap;

    fprintf(file, "\nLevels:\n");
    fprintf(file, "  lvl         name fsr fst1 fstN out1 outN fstmap\n");
    for (lvl = 1; lvl <= NLevels; lvl++) {
	fprintf(file, "  %3d %12s %3d %4d %4d",
		lvl, level_name(lvl) ? symbol_string(level_name(lvl)) : "(unnamed)",
		level_fsr(lvl), level_fst1(lvl), level_fstN(lvl));
	if (level_outN(lvl) < level_out1(lvl)) {
	    fprintf(file, "   --   --\n");
	}
	else {
	    fprintf(file, " %4d %4d", level_out1(lvl), level_outN(lvl));
	    fstmap = Ints + level_fstmap(lvl);
	    for (cat = level_out1(lvl); cat <= level_outN(lvl); cat++) {
		if (fstmap[cat]) {
		    fprintf(file, " %s:%d", symbol_string(cat), fstmap[cat]);
		}
	    }
	    terpri(file);
	}
    }
}


/**===================================================================
***
***				 REG
***
**/


int scan_reg (struct tokstream *stream)
{
    scan_att_decls(stream);
    while (scan_level(stream)) ;
}


int write_fsc (void *ignore, FILE *file)
{
    write_type(T_FSC, file);
    write_string(RegFilename != NULL ? RegFilename : "", file);
    write_regdict(file);
    write_fsas(file);
    write_amgr(LevelMgr, file);
}
    
void *read_fsc (FILE *file)
{
    TypeId t;

    read_type(t, file);
    if (t != T_FSC) error("read_fsc: not an fsc (bad cookie)");
    if (RegFilename != NULL) free(RegFilename);
    if (peek_char(file) == '\0') {
	RegFilename = NULL;
	read_char(file);
    }
    else RegFilename = read_string_alloc(file);
    read_regdict(file);
    read_fsas(file);
    LevelMgr = read_amgr(&Levels, file);
    return NULL;
}

int free_reg ()
{
    free_regdict();
    free_fsas();
    free_amgr(LevelMgr);
}


int load_reg (char *filename)
{
    struct tokstream *stream;

    if (!file_exists_p(filename)) return 0;
    stream = open_tokstream(filename);
    scan_reg(stream);
    if (RegFilename != NULL) free(RegFilename);
    RegFilename = strdup(filename);
    close_tokstream(stream);
    return 1;
}


load_fsc (char *filename)
{
    FILE *file;
    
    file = Fopen(filename, "r");
    read_fsc(file);
    fclose(file);
    if (FscFilename != filename) {
	if (FscFilename != NULL) free(FscFilename);
	FscFilename = strdup(filename);
    }
}


int save_fsc (char *filename)
{
    FILE *file;

    file = Fopen(filename, "w");
    write_fsc(NULL, file);
    fclose(file);
    if (FscFilename != NULL) free(FscFilename);
    FscFilename = strdup(filename);
}    


int dump_fsc (void *ignore, FILE *file)
{
    FsaId fsa;
    
    dump_regdict(file);
    terpri(file);
    dump_levels(file);

    for (fsa = 1; fsa <= NFsas; fsa++) {
	terpri(file);
	print_fsa(fsa, file);
    }
}


/**===================================================================
***
***				 MAIN
**/


compile_reg (char *reg_filename, char *fsc_filename)
{
    char buf[MAXFILENAME];
    FILE *file;

    if (has_suf(reg_filename, "fsc"))
      error("compile_reg: reg file not allowed to have .fsc suffix");
    if (!load_reg(reg_filename))
      error("compile_reg: can't find file %s", reg_filename);

    if (fsc_filename == NULL) {
	strcpy(buf, reg_filename);
	chsuf(buf, ".reg", ".fsc");
	fsc_filename = buf;
    }
    save_fsc(fsc_filename);
}
