#include <stdio.h>
#include <limits.h>  /* INT_MAX, INT_MIN */
#include <spa.h>
#include <rbtree.h>
#include <hash.h>
#include <regdict.h>
#include <regparse.h>



/**-------------------------------------------------------------------
***
***			      Reg Arrays
**/

int *Ints = NULL;
struct amgr *IntMgr;

typedef int ArrayId;


char *Chars = NULL;
struct amgr *CharMgr;


/**-------------------------------------------------------------------
***
***				 FSA
***
***  This structure contains all the permanent information about FSA's.
***  Information needed only during compilation is kept in a regex
***  structure.
***
***  The lists are indices of subarrays in the aarray Lists.
**/

typedef int FsaId;

struct fsa {
    SymbolId cat;
    int nlocals;     /* the number of locals */
    int ilocals;     /* the number of input locals (the rest are epsilon locals) */
    int nstates;     /* the number of states */
    SymbolId mincat;
    SymbolId maxcat;
    ArrayId cmap;    /* maps cat - cat0 to a local or -1 */
    ArrayId lmap;    /* maps local to cat */
    ArrayId emap;    /* maps local to type */
    ArrayId tm;      /* transition matrix */
    ArrayId finals;  /* array of cats */
};

struct fsa *Fsas = NULL;
struct amgr *FsaMgr;

#define NFsas (glength(FsaMgr))

FsaId CurrentFsa = 0;
SymbolId MinCat, MaxCat;
int NLocals, ILocals;
int *CMap, *LMap, *TM, *Finals, *TMap;
char *EMap;
struct amgr *TMapMgr;


int use_fsa (FsaId fsa);
int clear_fsa ();


#define new_fsa()    (gnew(FsaMgr))
#define NextFsa      (glength(FsaMgr)+1)

#define fsa_cat(i)       (Fsas[i].cat)
#define fsa_name(i)      (symbol_string(fsa_cat(i)))
#define fsa_nlocals(i)   (Fsas[i].nlocals)
#define fsa_ilocals(i)   (Fsas[i].ilocals)
#define fsa_nstates(i)   (Fsas[i].nstates)
#define fsa_mincat(i)    (Fsas[i].mincat)
#define fsa_maxcat(i)    (Fsas[i].maxcat)
#define fsa_cmap(i)      (Fsas[i].cmap)
#define fsa_lmap(i)      (Fsas[i].lmap)
#define fsa_emap(i)      (Fsas[i].emap)
#define fsa_tm(i)        (Fsas[i].tm)
#define fsa_finals(i)    (Fsas[i].finals)

int print_fsa (FsaId i, FILE *file);

SymbolId CatLocalTmp;
#define cat_local(cat) \
  (CatLocalTmp = (cat), \
   CatLocalTmp < MinCat || CatLocalTmp > MaxCat ? 0 \
   : CMap[CatLocalTmp])

#define local_cat(lcl)        (LMap[lcl])
#define elocal_type(lcl)      (EMap[lcl])
#define local_type(lcl)       (lcl <= ILocals ? INPUT : elocal_type(lcl))
#define tag_local(tag)        (TMap[tag])

char *local_string(int local)
{
    static char s[128];

    sprintf(s, "%s%s", tt_pstring(local_type(local)), symbol_string(local_cat(local)));
    return s;
}


/**--------------------------------------------------------------------
***
***				States
**/

int RegTraceTM = 0;

struct hash_table *States = NULL;
int StateHashAlloc = 1024;
int Nbits = 0, Nlits, Lit0, NStates = 0;

#define start_state         1
#define state_table(s)      (TM + (s) * NLocals)
#define state_next1(s,lcl)  (state_table(s)[lcl])

SymbolId StateNextTmp;
#define state_next(s,cat) \
  (StateNextTmp = (cat), \
   StateNextTmp < MinCat || StateNextTmp > MaxCat ? 0 : \
   state_next1(s, CMap[StateNextTmp]))

int StateFinalsTmp;
#define state_finals(s) \
  (StateFinalsTmp = Finals[s], \
   StateFinalsTmp ? Ints + StateFinalsTmp : NULL)



/**===================================================================
***
***			      Functions
**/



/**-------------------------------------------------------------------
***  
***				 Fsas
**/


int use_fsa (FsaId fsa)
{
    struct fsa *Fsa = &Fsas[fsa];

    CurrentFsa = fsa;
    NLocals = Fsa->nlocals;
    ILocals = Fsa->ilocals;
    MinCat = Fsa->mincat;
    MaxCat = Fsa->maxcat;
    CMap = Ints + Fsa->cmap;
    LMap = Ints + Fsa->lmap;
    EMap = Chars + Fsa->emap;
    if (TMap != NULL) gzero(TMapMgr);
    NStates = Fsa->nstates;
    TM = Ints + Fsa->tm;
    Finals = Ints + Fsa->finals;
}


int clear_fsa ()
{
    NLocals = ILocals = MinCat = MaxCat = 0;
    CMap = LMap = NULL;
    EMap = NULL;
    if (TMap != NULL) gzero(TMapMgr);
    NStates = 0;
    TM = NULL;
    Finals = NULL;
}
    

FsaId compile_regex (int rx)
{
    FsaId fsa;

    if (Fsas == NULL) init_fsas();
    fsa = new_fsa();
    fsa_cat(fsa) = regex_cat(rx);
    CurrentFsa = fsa;
    make_maps(fsa, rx);
    make_transition_matrix(fsa, rx);

    return fsa;
}


init_fsas ()
{
    if (Fsas == NULL) {
	FsaMgr = make_amgr(&Fsas, 100, sizeof(struct fsa));
	IntMgr = make_amgr(&Ints, 200, sizeof(int));
	CharMgr = make_amgr(&Chars, 100, sizeof(char));
    }
    else {
	gzero(FsaMgr);
	gzero(IntMgr);
	gzero(CharMgr);
    }
}


int write_fsas (FILE *file)
{
    write_amgr(FsaMgr, file);
    write_amgr(IntMgr, file);
    write_amgr(CharMgr, file);
}

int read_fsas (FILE *file)
{
    FsaMgr = read_amgr(&Fsas, file);
    IntMgr = read_amgr(&Ints, file);
    CharMgr = read_amgr(&Chars, file);
}


int free_fsas ()
{
    free_amgr(FsaMgr);
    free_amgr(IntMgr);
    free_amgr(CharMgr);
    if (TMap != NULL) free_amgr(TMapMgr);
    clear_fsa();
}


int print_fsa (FsaId fsa, FILE *file)
{
    FsaId prev = CurrentFsa;

    use_fsa(fsa);
    fprintf(file, "Fsa %d:\n", fsa);
    print_tm(file);
    use_fsa(prev);
}


int dump_fsa_stores (FILE *file)
{
    dump_fsas(file);
    terpri(file);
    dump_fsa_ints(file);
    terpri(file);
    dump_fsa_chars(file);
}


int dump_fsas (FILE *file)
{
    FsaId i;
    struct fsa *fsa;
    
    fprintf(file, "Fsas:\n");
    fprintf(file, "fsa cat nlocals ilocals nstates mincat maxcat cmap lmap emap  tm finals\n");
    for (i = 1; i <= NFsas; i++) {
	fsa = &Fsas[i];
	printf("%3d %3d %7d %7d %7d %6d %6d %4d %4d %4d %3d %6d\n",
	       i, fsa->cat, fsa->nlocals, fsa->ilocals, fsa->nstates, fsa->mincat, fsa->maxcat,
	       fsa->cmap, fsa->lmap, fsa->emap, fsa->tm, fsa->finals);

    }
}


int dump_fsa_ints (FILE *file)
{
    int i;
    
    fprintf(file, "Ints:\n");
    fprintf(file, "       1   2   3   4   5   6   7   8   9  10");
    for (i = 1; i <= glength(IntMgr); i++) {
	if (i%10 == 1) {
	    fprintf(file, "\n%3d:", i-1);
	}
	fprintf(file, " %3d", Ints[i]);
    }
}


int dump_fsa_chars (FILE *file)
{
    int i;
    
    fprintf(file, "Chars:\n");
    fprintf(file, "       1   2   3   4   5   6   7   8   9  10");
    for (i = 1; i <= glength(CharMgr); i++) {
	if (i%10 == 1) {
	    fprintf(file, "\n%3d:", i-1);
	}
	fprintf(file, " %3d", Chars[i]);
    }
    terpri(file);
}




/**-------------------------------------------------------------------
***
***			      MAKE MAPS
***
**/


make_maps (FsaId fsa, RegexId rx)
{
    int i, type, tag, cat, itags, *etags, with_actions;
    LiteralId lit;
    struct tree_state S;
    struct amgr *etag_mgr;
    
    MinCat = INT_MAX, MaxCat = INT_MIN;

    if (regex_litN(rx) < regex_lit1(rx)) error("make_maps: no lits");
    if (regex_is_deleted(rx)) error("make_maps: deleted regex");
    with_actions = regex_with_actions(rx);
    if (with_actions) {
	etag_mgr = make_amgr(&etags, 20, sizeof(int));
    }
    push_tstore(TodoTstore);

    NLocals = 0;
    ILocals = 0;
    itags = NIL;
    for (lit = regex_lit1(rx); lit <= regex_litN(rx); lit++) {
	tag = lit_tag(lit);
	type = tag_type(tag);
	if (type == INPUT || with_actions && (type&TAGBASE) == ACTION) {
	    if (type == INPUT) tinsert(tag, &itags);
	    else {
		gpush(&tag, etag_mgr);
		new_entry_p = 1;
	    }
	    if (new_entry_p) {
		NLocals++;
		if (type == INPUT) ILocals++;
		cat = tag_symbol(tag);
		if (cat < MinCat) MinCat = cat;
		if (cat > MaxCat) MaxCat = cat;
	    }
	}
    }
    if (MinCat > MaxCat) error("make_maps: no cats");

    fsa_nlocals(fsa) = NLocals;
    fsa_ilocals(fsa) = ILocals;
    fsa_mincat(fsa) = MinCat;
    fsa_maxcat(fsa) = MaxCat;

    fsa_cmap(fsa) = galloc(IntMgr, MaxCat - MinCat + 1) - MinCat;
    fsa_lmap(fsa) = galloc(IntMgr, NLocals) - 1;
    
    if (TMap == NULL) TMapMgr = make_amgr(&TMap, 100, sizeof(int));
    else gzero(TMapMgr);
    gsetlength(TMapMgr, NTags);

    /**  Here's a place where I had a very nasty bug.  Between the time  **/
    /**  you set CMap and the time you use it, you have to be absolutely **/
    /**  sure that Ints doesn't get moved (e.g., reallocated).  That     **/
    /**  means no galloc's on Ints, for example!                         **/
    CMap = Ints + fsa_cmap(fsa);
    LMap = Ints + fsa_lmap(fsa);

    tlist_start(&S, itags);
    for (i = 1; tlist_next(&tag, &S); i++) {
	TMap[tag] = i;
	CMap[tag_symbol(tag)] = i;
	LMap[i] = tag_symbol(tag);
    }

    pop_tstore();
    zero_tstore(TodoTstore);

    /**  Now the etags.  They have been interned left-to-right, so the sequence      **/
    /**  of tags corresponds to the sequence of literals.  We find the first right   **/
    /**  bracket and insert it, scan back over balanced brackets to find the         **/
    /**  corresponding left bracket, inserting each of the assignments at top        **/
    /**  level, then the left bracket.  Then we find the next right bracket or       **/
    /**  end-of-array and repeat.                                                    **/

    if (with_actions) {
	int rb, lb, t, d;

	fsa_emap(fsa) = galloc(CharMgr, NLocals-ILocals) - (1+ILocals);
	EMap = Chars + fsa_emap(fsa);

	if (ILocals + glength(etag_mgr) != NLocals)
	  error("make_maps: wrong number of elocals");

	rb = 0;
	i = 1 + ILocals;
	while (i <= NLocals) {

	    /* find right bracket */
	    while (++rb <= glength(etag_mgr)) {
		tag = etags[rb];
		if (tag_type(tag) == ACT_RIGHT) {
		    TMap[tag] = i;
		    /* rmap[i] = tag; */
		    cat = tag_symbol(tag);

		    /* this is wrong.  If there's an epsilon local */
		    /* with the same symbol as an input local, the */
		    /* input local gets clobbered in the cmap!     */
		    /* CMap[cat] = i; */

		    LMap[i] = cat;
		    EMap[i] = ACT_RIGHT;
		    i++;
		    break;
		}
	    }

	    /* find left bracket, collect assigns */
	    for (lb = rb-1, d = 0; lb > 0; lb--) {
		tag = etags[lb];
		t = tag_type(tag);
		if (t == ACT_ASSIGN) {
		    if (d == 0) {
			TMap[tag] = i;
			/* rmap[i] = tag; */
			cat = tag_symbol(tag);
			/* CMap[cat] = i; */
			LMap[i] = cat;
			EMap[i] = ACT_ASSIGN;
			i++;
		    }
		}
		else if (t == ACT_LEFT) {
		    if (d-- == 0) {
			TMap[tag] = i;
			/* rmap[i] = tag; */
			cat = tag_symbol(tag);
			/* CMap[cat] = i; */
			LMap[i] = cat;
			EMap[i] = ACT_LEFT;
			i++;
			break;
		    }
		}
		else d++;
	    }
	}
    }

    if (with_actions) free_amgr(etag_mgr);
}


int print_maps (FILE *file)
{
    int lcl;
    SymbolId cat;
    TagId tag;
  
    fprintf(file, "Maps:\n");
    fprintf(file, "  MinCat = %d, MaxCat = %d, NLocals = %d, ILocals = %d\n",
	    MinCat, MaxCat, NLocals, ILocals);
    terpri(file);
    fprintf(file, "CMap:\n");
    fprintf(file, "  cat local\n");
    for (cat = MinCat; cat <= MaxCat; cat++) {
	fprintf(file, "  %3d %5d %s\n", cat, CMap[cat], symbol_string(cat));
    }
    terpri(file);
    fprintf(file, "LMap:\n");
    fprintf(file, "  local cat\n");
    for (lcl = 1; lcl <= NLocals; lcl++) {
	cat = LMap[lcl];
	fprintf(file, "  %5d %3d %s\n", lcl, cat, cat == 0 ? "" : symbol_string(cat));
    }
    terpri(file);
    fprintf(file, "EMap:\n");
    fprintf(file, "  local type\n");
    for (lcl = 1 + ILocals; lcl <= NLocals; lcl++) {
	fprintf(file, "  %5d %4d %s\n", lcl, EMap[lcl], tt_pstring(EMap[lcl]));
    }
    if (TMap != NULL && glength(TMapMgr) > 0) {
	terpri(file);
	fprintf(file, "TMap:\n");
	fprintf(file, "  tag local\n");
	for (tag = 1; tag < NTags; tag++) {
	    fprintf(file, "  %3d %5d %s%s\n",
		    tag, TMap[tag], tt_pstring(tag_type(tag)), tag_string(tag));
	}
    }
}



/**-------------------------------------------------------------------
***
***				States
**/


init_states (RegexId rx)
{
    int oNbits = Nbits;
    
    Nlits = regex_litN(rx) - regex_lit1(rx) + 1;
    Lit0 = regex_lit1(rx);
    Nbits = ((Nlits+7)/8);
    if (States == NULL) States = make_hash_table(StateHashAlloc, Nbits, sizeof(int));
    else if (Nbits == oNbits) clrhash(States);
    else {
	free_hash_table(States);
	States = make_hash_table(StateHashAlloc, Nbits, sizeof(int));
    }
}
  

/**
***  lits is a tlist.  We convert it to a bit vector and use that as hash key.
***  States are numbered from 1 because 0 is the empty entry in the transition
***  matrix.
**/

int intern_state (int lits, struct tstore *tstore)
{
    int lit, *vp;
    struct tree_state list;
    char bits[Nbits];
  
    bzero(bits, Nbits);
    push_tstore(tstore);
    tlist_start(&list, lits);
    while (tlist_next(&lit, &list)) {
	lit -= Lit0;
	bits[lit/8] |= 1 << 7-lit%8;
    }
    pop_tstore();
    vp = needhash(bits, States);
    if (new_entry_p) *vp = States->length;
    return *vp;
}


/**-------------------------------------------------------------------
***
***				Trace
**/


#define Trace_start \
if (RegTraceTM) { \
    printf("\n"); \
    indent(stdout); \
    printf("Build Transition Matrix\n"); \
    push_indent(2); \
}
    

#define Trace_state \
if (RegTraceTM) { \
    indent(stdout); \
    printf("State %d, litset=", i); \
    print_Tlist(lits, tstore1, stdout); \
    printf("\n"); \
    push_indent(2); \
}


#define Trace_output \
if (RegTraceTM) { \
    indent(stdout); \
    printf("Lit %d, Tag %d=$%s, finals=", lit, tag, tag_string(tag)); \
    print_Tlist(finals, tstore1, stdout); \
    printf("\n"); \
}


#define Trace_input \
if (RegTraceTM) { \
    indent(stdout); \
    printf("Lit %d, Tag %d=%s%s, next_lits=", \
	   lit, tag, tt_pstring(tag_type(tag)), tag_string(tag)); \
    print_Tlist(next_lits, PointTstore, stdout); \
    printf(", a[%d] = ", local); \
    print_Tlist(a[local], tstore1, stdout); \
    printf("\n"); \
}


#define Trace_skip \
if (RegTraceTM) { \
    indent(stdout); \
    printf("Lit %d, Tag %d=%c%s, skipping\n", \
	   lit, tag, tt_pstring(tag_type(tag)), tag_string(tag)); \
}


#define Trace_next \
if (RegTraceTM) { \
    indent(stdout); \
    printf("Converting a[%d] from ", local); \
    print_Tlist(lits, tstore1, stdout); \
    printf(" to state %d", a[local]); \
    if (new_entry_p) printf(" (new)"); \
    printf("\n"); \
}


#define Trace_finals \
if (RegTraceTM) { \
    int TRACEi; \
    indent(stdout); \
    printf("Converting finals to array ["); \
    for (TRACEi = 0; TRACEi < nfinals; TRACEi++) { \
	if (TRACEi != 0) printf(" "); \
	printf("%d", fa[TRACEi]); \
    } \
    printf("]\n"); \
    pop_indent(); \
}


#define Trace_done \
if (RegTraceTM) { \
    printf("Done: %d states\n", NStates); \
    pop_indent(); \
}


/**-------------------------------------------------------------------
***
***				  TM
**/

/**  States are numbered from 1.  state_litsets is indexed by state-1.  **/

static int *StateLitsets = NULL, *M, *FinalLists;
static struct amgr *StateLitsetMgr, *MMgr, *FinalListMgr;


make_transition_matrix (FsaId fsa, RegexId rx)
{
    int *a, i, j, lit, lits, *litsp, next_lits, state, tag, local, locals, next,
    cat, finals, nfinals, flist, *fa, rp, startp, endp, with_actions;
    struct tree_state S;
    struct tstore *tstore1, *tstore2;
  
    init_states(rx);

    startp = regex_lp(rx);
    endp = regex_rp(rx);
    with_actions = regex_with_actions(rx);

    if (intern_state(point_lits(startp), PointTstore) != 1)
      error("make_transition_matrix: start state not one");
    if (StateLitsets == NULL) {
	StateLitsetMgr = make_amgr(&StateLitsets, 100, sizeof(int));
	MMgr = make_amgr(&M, 100, sizeof(int));
	FinalListMgr = make_amgr(&FinalLists, 50, sizeof(int *));
    }
    else {
	gzero(StateLitsetMgr);
	gzero(MMgr);
	gzero(FinalListMgr);
    }
    tstore1 = TodoTstore;
    tstore2 = AuxTstore;
    lits = NIL;
    Tlist_start(&S, point_lits(startp), PointTstore);
    while (Tlist_next(&lit, &S, PointTstore))
      Tinsert(lit, &lits, tstore1);
    gpush(&lits, StateLitsetMgr);
    Trace_start;

    /**  Go through states.  This vector will grow as new states are interned.  **/
    for (i = 1; i <= States->length; i++) {
    
	a = M + galloc(MMgr, NLocals) - 1;
	finals = NIL;
	nfinals = 0;
	lits = StateLitsets[i];
	Trace_state;

	/**  Go through lits for this state.  **/
	zero_tstore(tstore2), locals = NIL;
	Tlist_start(&S, lits, tstore1);
	while (Tlist_next(&lit, &S, tstore1)) {
	    tag = lit_tag(lit);

	    /**  Output literal.  **/
	    if (tag_type(tag) == OUTPUT) {
		Tinsert(tag_symbol(tag), &finals, tstore1);
		nfinals++;
		Trace_output;
	    }

	    /**  Input or epsilon literal.  **/
	    else if (with_actions || tag_type(tag) == INPUT) {
		local = tag_local(tag);
		if (local > NLocals) error("local out of bounds");
		Tinsert(local, &locals, tstore2);
		next_lits = point_lits(lit_rp(lit));
		Tnunion(&a[local], tstore1, next_lits, PointTstore);
		Trace_input;
	    }

	    else Trace_skip;
	}

	/**  Next-state table contains litsets; convert them to states.  **/
	Tlist_start(&S, locals, tstore2);
	while (Tlist_next(&local, &S, tstore2)) {
	    lits = a[local];
	    a[local] = intern_state(lits, tstore1);
	    if (new_entry_p) gpush(&lits, StateLitsetMgr);
	    Trace_next;
	}

	/**  Construct list of finals for this state.  **/
	if (nfinals > 0) {
	    flist = galloc(IntMgr, nfinals+1);
	    fa = Ints + flist;
	    push_tstore(tstore1);
	    tlist_start(&S, finals);
	    j = 0;
	    while (tlist_next(&cat, &S)) {
		fa[j++] = cat;
	    }
	    pop_tstore();
	    if (j != nfinals) error("bad count for finals");
	    fa[j] = 0;
	}
	else flist = 0;
	gpush(&flist, FinalListMgr);
	Trace_finals;
    }

    /**  Done all states.  **/
    NStates = States->length;
    Trace_done;

    fsa_nstates(fsa) = NStates;

    /**  We want a = tm + 1*NLocals + 1  **/
    /**  hence Ints + i = Ints + tm + 1*NLocals + 1  **/
    /**  hence tm = i - NLocals - 1  **/
    
    i = galloc(IntMgr, NStates * NLocals);
    fsa_tm(fsa) = i - NLocals - 1;
    TM = Ints + fsa_tm(fsa);
    bcopy(&M[1], &Ints[i], NStates * NLocals * sizeof(int));

    i = galloc(IntMgr, NStates);
    fsa_finals(fsa) = i - 1;
    Finals = Ints + i - 1;
    bcopy(&FinalLists[1], &Ints[i], NStates * sizeof(int));

    zero_tstore(tstore1);
    zero_tstore(tstore2);
}


print_tm (FILE *file)
{
    int s, i, *a, *f, cat;
    struct tree_state S;
    char *string;

    fprintf(file, "    ");
    for (i = 1; i <= NLocals; i++) {
	fprintf(file, " %4d", i);
	if (i == ILocals) fprintf(file, " :");
    }
    terpri(file);
    fprintf(file, "    ");
    for (i = 1; i <= NLocals; i++) {
	string = local_string(i);
	string[4] = 0;
	fprintf(file, " %4s", string);
	if (i == ILocals) fprintf(file, " :");
    }
    fprintf(file, "  | Final\n");
    indent(file);

    for (s = 1; s <= NStates; s++) {
	indent(file);
	fprintf(file, "%3d:", s);
	for (i = 1; i <= NLocals; i++) {
	    fprintf(file, " %4d", state_next1(s,i));
	    if (i == ILocals) fprintf(file, " :");
	}
	print_string("  |", file);
	f = state_finals(s);
	if (f != NULL) {
	    for (; *f; f++) {
		fprintf(file, " %s", symbol_string(*f));
	    }
	}
	fprintf(file, "\n");
    }
}

