#include <stdio.h>
#include <spa.h>
#include <amgr.h>
#include <hash.h>
#include <rbtree.h>
#include <dict.h>
#include "regdict.h"
#include "spatok.h"



/* extending type list from spatok.h */

#define TERM  NTOKTYPES
#define SEQ   NTOKTYPES+1
#define ALT   NTOKTYPES+2
#define EXP   NTOKTYPES+3

#define NREGTYPES  NTOKTYPES+4

static char *NodeTypeNames[] = { "Term", "Seq", "Alt", "Exp" };

int RegTraceRX = 0;
int PrintCompileMessages = 1;

#define STACKSIZE 128
#define ESTACKSIZE 32

char *type_name (int t);



typedef int RegexId;
typedef int PointId;
typedef int LiteralId;
typedef int ListId;


/**===================================================================
***
***				Regex
**/

struct regex {
    SymbolId cat;
    int flags;
    PointId lp;
    PointId rp;
    PointId pt1;
    PointId ptN;
    LiteralId lit1;
    LiteralId litN;
};

struct regex *Regices = NULL;
struct amgr *RegexMgr;

#define new_regex()    (gnew(RegexMgr))

#define regex_cat(i)   (Regices[i].cat)
#define regex_flags(i) (Regices[i].flags)
#define regex_lp(i)    (Regices[i].lp)
#define regex_rp(i)    (Regices[i].rp)
#define regex_pt1(i)   (Regices[i].pt1)
#define regex_ptN(i)   (Regices[i].ptN)
#define regex_lit1(i)  (Regices[i].lit1)
#define regex_litN(i)  (Regices[i].litN)
#define regex_fsa(i)   (Regices[i].fsa)

#define NextRegex      (glength(RegexMgr) + 1)
#define NRegices       (glength(RegexMgr))

#define RX_MACRO       1
#define RX_NO_ACTIONS  2
#define RX_DELETED     4

#define regex_is_macro(rx)     (GET(regex_flags(rx), RX_MACRO))
#define regex_with_actions(rx) (!GET(regex_flags(rx), RX_NO_ACTIONS))
#define regex_is_deleted(rx)   (GET(regex_flags(rx), RX_DELETED))

int init_regices ();

RegexId scan_regex (struct tokstream *file);
RegexId find_regex (SymbolId cat);
int print_regex (int r, FILE *file);


/**-------------------------------------------------------------------
***
***				Points
**/

struct point {
    ListId lits;
    PointId to1;
    PointId to2;
};

struct point *Points = NULL;
struct amgr *PointMgr;

struct tstore *PointTstore;
struct tstore *TodoTstore;
struct tstore *AuxTstore;

#define new_point()    (gnew(PointMgr))
#define NextPoint      (glength(PointMgr)+1)
#define NPoints        (glength(PointMgr))

#define point_lits(p)  (Points[p].lits)
#define point_to1(p)   (Points[p].to1)
#define point_to2(p)   (Points[p].to2)

#define print_points(file) (print_point_range(1, NPoints, (file)))

#define print_regex_points(file) \
  (print_point_range(Regex->pt1, Regex->ptN, (file)))



/**-------------------------------------------------------------------
***
***			       Literals
**/

struct literal {
    TagId tag;
    PointId rp;
};

struct literal *Literals = NULL;
struct amgr *LiteralMgr;

#define new_literal()   (gnew(LiteralMgr))
#define NextLiteral     (glength(LiteralMgr)+1)
#define NLiterals       (glength(LiteralMgr))

#define lit_tag(lit)    (Literals[lit].tag)
#define lit_rp(lit)     (Literals[lit].rp)
#define lit_string(lit) (tag_string(lit_tag(lit)))

#define print_literals(file) (print_literal_range(1,NLiterals,(file)))


/**-------------------------------------------------------------------
***
***				Nodes
**/

struct node {
    int t;
    SymbolId v;
};


/**===================================================================
***
***			      Functions
**/


char *type_name (int t)
{

    if (t >= NREGTYPES) return "(bad type)";
    else if (t < NTOKTYPES) return toktype_string(t);
    else return NodeTypeNames[t-NTOKTYPES];
}


int init_regices ()
{
    if (Regices == NULL) {
	RegexMgr = make_amgr(&Regices, 100, sizeof(struct regex));
	PointMgr = make_amgr(&Points, 200, sizeof(struct point));
	LiteralMgr = make_amgr(&Literals, 200, sizeof(struct literal));
	PointTstore = make_tstore(200);
	TodoTstore = make_tstore(200);
	AuxTstore = make_tstore(200);
    }
    else {
	clear_tags();
	gzero(RegexMgr);
	gzero(PointMgr);
	gzero(LiteralMgr);
	zero_tstore(PointTstore);
	zero_tstore(TodoTstore);
	zero_tstore(AuxTstore);
    }
}


RegexId find_regex (SymbolId cat)
{
    int i;
    
    for (i = 1; i <= NRegices; i++) {
	if (Regices[i].cat == cat) {
	    return i;
	}
    }
    return 0;
}


/**-------------------------------------------------------------------
***
***			      scan_regex
**/


#define RegTraceSyntax 1
#define RegTraceDump   2

#define Trace_Stack \
if (RegTraceRX == RegTraceDump) { \
    printf("\n  "); \
    for (i = 0; i <= top; i++) { \
	print_node(&stack[i], stdout); \
	printf(" "); \
    } \
    printf("/ "); \
    print_node(&la, stdout); \
    printf("\n"); \
}


#define Trace_Macro \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	indent(stdout); \
	printf("[Term %d,%d]  ->  Macro %s\n", lp, rp, symbol_string(stack[top].v)); \
    } \
    else printf("  Expand macro %s -> [term %d %d]\n", \
	        symbol_string(stack[top].v), lp, rp); \
}


#define Trace_Term \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	indent(stdout); \
	printf("[Term %d %d]  ->  %s%s (lit %d)\n", \
	       lp, rp, tt_pstring(tag_type(tag)), tag_string(tag), lit); \
    } \
    else if (t == STRING) \
      printf("  Reduce [string %s] to [term %d %d], lit=%d, tag=%d (%s:%s)\n", \
	     lit_string(lit), lp, rp, lit, tag, tag_string(tag), tt_string(tag_type(tag))); \
    else \
      printf("  Shift, Reduce COLON/LBRACKET [string %s] to [term %d %d], lit=%d, tag=%d (%s:%s)\n", \
	     lit_string(lit), lp, rp, lit, tag, tag_string(tag), tt_string(tag_type(tag))); \
}


#define Trace_ShEq \
if (RegTraceRX == RegTraceDump) { \
    printf("  Shift EQUALS\n"); \
}


#define Trace_ReEq \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	printf("[%s %d %d]  ->  [Term %d %d] equals [%s %d %d]\n", \
	       type_name(t), pl, pr, lcl, lcr, type_name(t), rcl, rcr); \
    } \
    else printf("  Reduce [term %d %d] EQUALS [%s %d %d] to [%s %d %d]\n", \
		lcl, lcr, type_name(t), rcl, rcr, type_name(t), pl, pr); \
}



#define Trace_Star \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	printf("[Term %d %d]  ->  [Term %d %d] %s\n", \
	       pl, pr, lcl, lcr, type_name(la.t)); \
    } \
    else printf("  Shift, Reduce [term %d %d] QMARK/STAR/PLUS to [term %d %d]\n", \
		lcl, lcr, pl, pr); \
}



#define Trace_Seq \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	printf("[Seq %d %d]  ->  [%s %d %d] [Term %d %d]\n", \
	       pl, pr, type_name(s), lcl, lcr, rcl, rcr); \
    } \
    else printf("  Reduce [seq/term %d %d] [term %d %d] to [seq %d %d]\n", \
	   lcl, lcr, rcl, rcr, pl, pr); \
}



#define Trace_Lparen \
if (RegTraceRX == RegTraceDump) { \
    printf("  Shift STRING/LPAREN\n"); \
}


#define Trace_Rparen \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	printf("[Term %d %d]  ->  lparen [%s %d %d] rparen\n", \
	       stack[top].v, stack[top].v+1, \
	       type_name(t), stack[top].v, stack[top].v+1); \
    } \
    else printf("  Shift, Reduce TERM -> LPAREN ALT/SEQ/TERM RPAREN\n"); \
}


#define Trace_Alt \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	printf("[Alt %d %d]  ->  [%s %d %d] vbar [%s %d %d]\n", \
	       pl, pr, type_name(r), lcl, lcr, \
	       type_name(t), rcl, rcr); \
    } \
    else printf("  Reduce [alt/seq/term %d %d] VBAR [seq/term %d %d] to [alt %d %d]\n", \
	   lcl, lcr, rcl, rcr, pl, pr); \
}


#define Trace_Vbar \
if (RegTraceRX == RegTraceDump) printf("  Shift VBAR\n")

      
#define Trace_Accept \
if (RegTraceRX) { \
    if (RegTraceRX == RegTraceSyntax) { \
	printf("[Regex %d %d]  ->  [%s %d %d] eos\n", \
	       lp, rp, type_name(t), lp, rp);\
    } \
    else { \
	printf("  Accept %s, lit=%d, tag=%d\n", \
	       symbol_string(regex_cat(rx)), eos, lit_tag(eos)); \
	print_regex(rx, stdout); \
	print_points(stdout); \
	print_literals(stdout); \
    } \
}



#define NEXT_TOKEN \
(la.t = scan_token(file), \
 la.t == STRING ? (la.v = intern(file->token)) : 0)

RegexId scan_regex (struct tokstream *file)
{
    struct node stack[STACKSIZE], la;
    int estack[ESTACKSIZE];
    ListId todo;
    int top, r, s, t, i, etop = -1;
    PointId lp, rp, pl, pr, lcl, lcr, rcl, rcr;
    RegexId rx;
    SymbolId lhs;
    TType tt;
    TagId tag;
    LiteralId lit;

    if (Regices == NULL) init_regices();
    todo = NIL;

    if ((tt = peek_token(file)) == COLON || tt == EOF) return 0;
    tt = scan_token(file);
    if (tt != STRING) {
	error("%s, line %d, token %s: expecting lhs",
	      file->filename, file->line, toktype_string(tt));
    }
    if (PrintCompileMessages) printf("Pattern %s\n", file->token);
    lhs = intern(file->token);
    if (find_regex(lhs)) {
	error("Regex %s defined twice", file->token);
    }
    tt = scan_token(file);
    if (!(tt == ARROW || tt == EQUALS)) {
	error("%s, line %d, token %s: expecting arrow/equals",
	      file->filename, file->line, toktype_string(tt));
    }
    
    rx = new_regex();
    regex_cat(rx) = lhs;
    regex_lit1(rx) = NextLiteral;
    regex_pt1(rx) = NextPoint;

    if (tt == EQUALS) SET(regex_flags(rx), RX_MACRO);

    NEXT_TOKEN;  
    stack[0] = la;
    top = 0;
    la.t = 0;

    while (1) {
	t = stack[top].t;
	s = (top > 0 ? stack[top-1].t : 0);
	r = (top > 1 ? stack[top-2].t : 0);
	if (la.t == 0) NEXT_TOKEN;
	Trace_Stack;


	/**  macro  **/

	if (t == STRING && expand_macro(stack[top].v, &lp, &rp)) {
	    Trace_Macro;
	    stack[top].t = TERM;
	    stack[top].v = lp;
	    Tinsert(lp, &todo, TodoTstore);
	}


	/**  string / equals => [term A string] / equals  **/
	/**  string          => [term string]             **/
	/**  lb+rb / string  => [term E string]           **/

	else if (t == STRING || t == LBRACKET || t == RBRACKET) {
	    lit = new_literal();
	    lp = new_point(), rp = new_point();
	    if (t == STRING && la.t == EQUALS)
	      tag = intern_tag(stack[top].v, ACT_ASSIGN);
	    else if (t == STRING)
	      tag = intern_tag(stack[top].v, INPUT);
	    else if (t == LBRACKET) {
		if (la.t != STRING)
		  error("scan_regex: syntax error (missing string after lbracket)");
		tag = intern_tag(la.v, ACT_LEFT);
		la.t = 0;
		if (etop >= ESTACKSIZE) error("scan_regex: estack overflow");
		estack[++etop] = la.v;
	    }
	    else {
		if (etop < 0) error("scan_regex: syntax error (unmatched brackets)");
		tag = intern_tag(estack[etop--], ACT_RIGHT);
	    }
	    lit_tag(lit) = tag;
	    lit_rp(lit) = rp;
	    Tinsert(lit, &point_lits(lp), PointTstore);
	    Tinsert(lp, &todo, TodoTstore);
	    stack[top].t = TERM;
	    stack[top].v = lp;
	    Trace_Term;
	}


	/**  term / equals   => term equals /  **/

	else if (t == TERM && (la.t == EQUALS)) {
	    stack[++top].t = la.t;
	    stack[top].v = la.v;
	    la.t = 0;
	    Trace_ShEq;
	}


	/**  term equals term+seq+alt  => [term+seq+alt term term+seq+alt]  **/

	else if (r == TERM &&
		 s == EQUALS &&
		 (t == TERM || t == SEQ || t == ALT)) {
	    pl = new_point();
	    pr = new_point();
	    lcl = stack[top-2].v;
	    lcr = lcl+1;
	    rcl = stack[top].v;
	    rcr = rcl+1;
	    point_to1(lcl) = pl;
	    point_to1(rcl) = lcr;
	    point_to1(pr) = rcr;
	    top -= 2;
	    stack[top].t = t;
	    stack[top].v = pl;
	    Trace_ReEq;
	}


	/**  term / qmark+star+plus  =>  [qmark+star+plus term]  **/

	else if (t == TERM && (la.t == QMARK || la.t == STAR || la.t == PLUS)) {
	    pl = new_point();
	    pr = new_point();
	    lcl = stack[top].v;
	    lcr = lcl+1;
	    point_to1(lcl) = pl;
	    point_to1(pr) = lcr;
	    if (la.t == QMARK || la.t == STAR)
	      point_to2(pr) = pl;
	    if (la.t == PLUS || la.t == STAR)
	      point_to2(lcl) = lcr;
	    stack[top].v = pl;
	    Trace_Star;
	    la.t = 0;
	}
      

	/**  seq+term term   =>  [seq seq+term term]  **/

	else if (t == TERM && (s == TERM || s == SEQ)) {
	    pl = new_point();
	    pr = new_point();
	    lcl = stack[top-1].v;
	    lcr = lcl+1;
	    rcl = stack[top].v;
	    rcr = rcl+1;
	    point_to1(lcl) = pl;
	    point_to1(rcl) = lcr;
	    point_to1(pr) = rcr;
	    top--;
	    stack[top].t = SEQ;
	    stack[top].v = pl;
	    Trace_Seq;
	}


	/**  seq+term+vbar+lparen+equals / string+lparen  => shift  **/

	else if ((t == SEQ || t == TERM || t == VBAR || t == LPAREN || t == EQUALS)
		 && (la.t == STRING || la.t == LPAREN
		     || la.t == LBRACKET || la.t == RBRACKET)) {
	    top++;
	    stack[top].t = la.t;
	    stack[top].v = la.v;
	    la.t = 0;
	    Trace_Lparen;
	}


	/**  lparen alt+seq+term / rparen  =>  [term alt+seq+term]  **/

	else if ((t == ALT || t == SEQ || t == TERM) &&
		 s == LPAREN && la.t == RPAREN) {
	    stack[top-1].t = TERM;
	    stack[top-1].v = stack[top].v;
	    top--;
	    la.t = 0;
	    Trace_Rparen;
	}

	/**  alt+seq+term vbar seq+term   =>  [alt 1 3]  **/
	else if (s == VBAR
		 && (r == ALT || r == SEQ || r == TERM)
		 && (t == SEQ || t == TERM)) {
	    pl = new_point();
	    pr = new_point();
	    lcl = stack[top-2].v;
	    lcr = lcl+1;
	    rcl = stack[top].v;
	    rcr = rcl+1;
	    point_to1(lcl) = pl;
	    point_to1(rcl) = pl;
	    point_to1(pr) = lcr;
	    point_to2(pr) = rcr;
	    top -= 2;
	    stack[top].t = ALT;
	    stack[top].v = pl;
	    Trace_Alt;
	}


	/**  alt+seq+term / vbar: shift  **/

	else if (la.t == VBAR && (t == ALT || t == SEQ || t == TERM)) {
	    top++;
	    stack[top].t = la.t;
	    stack[top].v = la.v;
	    la.t = 0;
	    Trace_Vbar;
	}


	/**  Accept  **/

	else if (top == 0 && la.t == SEMICOLON
		 && (t == TERM || t == SEQ || t == ALT)) {
	    LiteralId eos = new_literal();
	    lp = stack[top].v;
	    rp = lp+1;
	    lit_tag(eos) = intern_tag(regex_cat(rx), OUTPUT);
	    Tinsert(eos, &point_lits(rp), PointTstore);
	    Tinsert(rp, &todo, TodoTstore);
	    regex_lp(rx) = lp;
	    regex_rp(rx) = rp;
	    regex_ptN(rx) = NPoints;
	    regex_litN(rx) = NLiterals;
	    propagate(todo);
	    Trace_Accept;
	    return rx;
	}

	else if (top == 0) {
	    error("%s, line %d, token %s: syntax error in regex (missing semicolon?)",
		  file->filename, file->line, toktype_string(la.t));
	}
	else {
	    error("%s, line %d, token %s: syntax error in regex",
		  file->filename, file->line, toktype_string(la.t));
	}
    }
}

    
/**-------------------------------------------------------------------
***
***  Expand Macro
***
***  Copy literals and points.  Returns 1 if macro found, 0 if not found.
***
***  Literals only have a tag and rp.  For the macro, we know pt0 and npts.
***  We also know pt0 for the copy.  The i-th point of the macro maps to
***  the i-th point of the copy.  So we can translate a literal's rp just
***  by adding the difference between the copy's pt0 and the macro's pt0.
***
***  Likewise we can translate a lit index by adding the difference between
***  the copy's lit0 and the original's lit0.  We do that to construct a new
***  litset for the copy of a point.  We also have to translate the point's
***  to1 and to2.
**/


int expand_macro (SymbolId cat, PointId *lpp, PointId *rpp)
{
    int i, ptdiff, litdiff, oldp, newp, oldlit, newlit, to, lit;
    struct regex *r;
    struct tree_state S;

    i = find_regex(cat);
    if (i == 0) return 0;
    r = &Regices[i];

    ptdiff = NextPoint - r->pt1;
    litdiff = NextLiteral - r->lit1;

    push_tstore(PointTstore);
    for (oldp = r->pt1; oldp <= r->ptN; oldp++) {
	newp = new_point();
	if (to = point_to1(oldp))
	  point_to1(newp) = to + ptdiff;
	if (to = point_to2(oldp))
	  point_to2(newp) = to + ptdiff;
	tlist_start(&S, point_lits(oldp));
	while (tlist_next(&lit, &S)) {
	    if (tag_type(lit_tag(lit)) != OUTPUT)
	      tinsert(lit + litdiff, &point_lits(newp));
	}
    }

    for (oldlit = r->lit1; oldlit <= r->litN; oldlit++) {
	newlit = new_literal();
	lit_tag(newlit) = lit_tag(oldlit);
	lit_rp(newlit) = lit_rp(oldlit) + ptdiff;
    }
      
    pop_tstore();

    if (r->rp != r->lp + 1) error("copy_macro_regex: regex with bad points");

    *lpp = r->lp + ptdiff;
    *rpp = r->rp + ptdiff;

    return 1;
}


/**-------------------------------------------------------------------
***
***			      PROPAGATE
**/


int propagate (ListId todo)
{
    int next, p, lit, to1, to2;
    struct tree_state pts, lits;
    struct tstore *tstore1, *tstore2, *tmp;

    tstore1 = TodoTstore;
    tstore2 = AuxTstore;

    for (next = NIL, zero_tstore(tstore2);
	 todo != NIL;
	 todo = next, next = NIL,
	 tmp = tstore1, tstore1 = tstore2, tstore2 = tmp,
	 zero_tstore(tstore2)) {

	Tlist_start(&pts, todo, tstore1);
	while (Tlist_next(&p, &pts, tstore1)) {
	    to1 = point_to1(p), to2 = point_to2(p);
	    Tlist_start(&lits, point_lits(p), PointTstore);
	    while (Tlist_next(&lit, &lits, PointTstore)) {
		if (to1) {
		    Tinsert(lit, &point_lits(to1), PointTstore);
		    if (new_entry_p) {
			Tinsert(to1, &next, tstore2);
		    }
		}
		if (to2) {
		    Tinsert(lit, &point_lits(to2), PointTstore);
		    if (new_entry_p) {
			Tinsert(to2, &next, tstore2);
		    }
		}
	    }
	}
    }
    zero_tstore(tstore1);
}



/**-------------------------------------------------------------------
***
***			   Delete Epsilons
**/


int delete_actions (RegexId rx)
{
    LiteralId lit;
    TagId tag;
    PointId p;
    ListId todo;

    push_tstore(TodoTstore);
    todo = NIL;
    for (lit = regex_lit1(rx); lit <= regex_litN(rx); lit++) {
	tag = lit_tag(lit);
	if ((tag_type(tag) & TAGBASE) == ACTION) {
	    p = lit_rp(lit);
	    point_to1(p) = p-1;
	    tinsert(p, &todo);
	}
    }
    pop_tstore();
    SET(regex_flags(rx), RX_NO_ACTIONS);
    propagate(todo);
}


/**-------------------------------------------------------------------
***
***			   COMBINE REGICES
***
***  This puts propagation links across all epsilon literals and
***  repropagates, then constructs new maps and a new transition matrix.
***
***  We reset Literals after each level, so we can find all the literals
***  for all the regices simply by walking through Literals.
***
***  The lp of a literal is one less than the rp of the literal.
***  Every epsilon literal does have a rp.  (Output literals don't, but
***  we don't care about output literals here.)
**/


RegexId combine_regices (SymbolId cat, RegexId rx1, RegexId rxN)
{
    LiteralId lit;
    TagId tag;
    PointId p, lp, rp;
    RegexId rx, r;
    ListId todo;

    if (rxN < rx1) error("combine_regices: rxN < rx1");
    rx = new_regex();
    /* I'm not sure that this matters anymore */
    if (rx != rxN + 1)
      error("combine_regices: attempt to create regex with noncontiguous points");
    lp = new_point();
    rp = new_point();

    regex_cat(rx) = cat;
    regex_lit1(rx) = regex_lit1(rx1);
    regex_litN(rx) = regex_litN(rxN);
    regex_lp(rx) = lp;
    regex_rp(rx) = rp;
    regex_pt1(rx) = regex_pt1(rx1);
    regex_ptN(rx) = rp;

    push_tstore(TodoTstore);
    todo = NIL;
    for (r = rx1; r <= rxN; r++) {
	if (!regex_is_macro(r)) {
	    if (GET(regex_flags(r), RX_DELETED)) error("combine_regices: deleted regex");
	    p = regex_lp(r);
	    point_to1(p) = lp;
	    tinsert(p, &todo);
	    SET(regex_flags(r), RX_DELETED);
	}
    }
    pop_tstore();
    propagate(todo);

    return rx;
}



/**-------------------------------------------------------------------
***
***  Print Regex
**/


print_regex (RegexId r, FILE *file)
{
    int i;

    fprintf(file, "Regex %s:\n", regex_cat(r) > 0 ? symbol_string(regex_cat(r)) : "(anon)");
    push_indent(2);

    indent(file);
    fprintf(file, "cat=%s(%d), flags=%d, lp=%d, rp=%d\n",
	    symbol_string(regex_cat(r)), regex_cat(r), regex_flags(r),
	    regex_lp(r), regex_rp(r));

    indent(file);
    fprintf(file, "pts=[%d %d], lits=[%d %d]\n",
	    regex_pt1(r), regex_ptN(r),
	    regex_lit1(r), regex_litN(r));

    pop_indent();
    
    terpri(file);
    print_point_range(regex_pt1(r), regex_ptN(r), file);
    terpri(file);
    print_literal_range(regex_lit1(r), regex_litN(r), file);
}


int dump_regices (FILE *file)
{
    RegexId rx;
    
    fprintf(file, "Regices:\n");
    fprintf(file, "  rx          cat flags lp rp pt1 ptN lit1 litN\n");
    for (rx = 1; rx <= NRegices; rx++) {
	fprintf(file, "  %2d %12s %5d %2d %2d %3d %3d %4d %4d\n",
		rx, regex_cat(rx) ? symbol_string(regex_cat(rx)) : "(unnamed)",
		regex_flags(rx), regex_lp(rx), regex_rp(rx), regex_pt1(rx), regex_ptN(rx),
		regex_lit1(rx), regex_litN(rx));
    }
}


print_point_range (PointId start, PointId end, FILE *file)
{
    PointId p, to1, to2;
    int lit, lits;
    struct tree_state S;

    indent(file);
    fprintf(file, "Points:\n");
    fprintf(file, " pt to1 to2 lits\n");
    for (p = start; p <= end; p++) {
	fprintf(file, "%3d", p);
	lits = point_lits(p), to1 = point_to1(p), to2 = point_to2(p);
	if (to1) fprintf(file, " %3d", to1);
	else fprintf(file, "    ");
	if (to2) fprintf(file, " %3d ", to2);
	else fprintf(file, "     ");
	if (lits) print_Tlist(lits, PointTstore, file);
	fprintf(file, "\n");
    }
}

print_literal (int lit, FILE *file)
{
  int tag = lit_tag(lit);

  fprintf(file, "<lit-%d, tag=%d (%s:%s), rp=%d>",
	  lit, tag, tag_string(tag), tt_string(tag_type(tag)), lit_rp(lit));
}

print_literal_range (int start, int end, FILE *file)
{
    LiteralId lit;
    TagId tag;

    fprintf(file, "Literals:\n");
    fprintf(file, "lit  rp tag\n");
    for (lit = start; lit <= end; lit++) {
	tag = lit_tag(lit);
	fprintf(file, "%3d %3d %3d %s%s\n", lit, lit_rp(lit),
		tag, tt_pstring(tag_type(tag)), tag_string(tag));
    }
}


print_node (struct node *n, FILE *file)
{
  fprintf(file, "[%s", type_name(n->t));
  if (n->t == STRING)
    fprintf(file, " %s", symbol_string(n->v));
  else if (n->t == TERM || n->t == SEQ || n->t == ALT)
    fprintf(file, " %d %d", n->v, n->v+1);
  fprintf(file, "]");
}




