#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>
#include "spa.h"
#include "spamap.h"



/**
***  Open a memory-mapped file.  The modes this pays attention to are:
***
***      r     open an existing map and read it in; size ignored
***      w     create a new map with associated file
**/

struct map *open_map (char *filename, char *mode, int size)
{
    if (mode[0] == 'r') return read_map(filename);
    else if (mode[0]== 'w') return create_map(filename, size);
    else error("open_map: unrecognized mode %s", mode);
}



/**  Todo: mmap /dev/zero and write from there instead of zeroing a byte at
***  a time.  */


struct map *create_map (char *filename, int size)
{
    struct map *map;
    int fd;
    void *data;
    
    allocate(map, 1, struct map);
    if (filename == NULL) {
	map->fd = -1;
	map->alloc = size;
	allocate(map->data, size, char);
    } else {
	fd = open(filename,
		  O_CREAT | O_TRUNC | O_RDWR,
		  S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
	if (fd < 0)
	  error("create_map: Failed to open %s", filename);
	ftruncate(fd, size);
	data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (data < (void *) 0)
	  error("create_map: Failed to map");
	map->fd = fd;
	map->alloc = size;
	map->data = data;
    }
    return map;
}


struct map *read_map (char *filename)
{
    struct map *map;
    struct stat stat_info;
    int size, fd;
    void *data;
    
    allocate(map, 1, struct map);
    if (filename == NULL) error("read_map: Need a filename");
    if (stat(filename, &stat_info) < 0)
      error("read_map: Unable to open %s", filename);
    if (stat_info.st_size > INT_MAX)
      error("open_map: Sorry, this file's too big (%s)", filename);
    size = stat_info.st_size;
    fd = open(filename, O_RDWR);
    if (fd < 0)
      error("open_map: Unable to open %s", filename);
    data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (data < (void *) 0) {
	fprintf(stderr, "read_map: Failed to map: ");
	perror("");
	exit(1);
    }
    map->fd = fd;
    map->alloc = size;
    map->data = data;

    return map;
}


int close_map (struct map *map)
{
    if (map->fd < 0) {
	free(map->data);
    } else {
	close(map->fd);
	munmap(map->data, map->alloc);
    }
    free(map);
}


int resize_map (struct map *map, int size)
{
    if (map->fd < 0) {
	if (size != map->alloc) {
	    reallocate(map->data, map->alloc, size, char);
	}
    } else {
	if (size != map->alloc) {
	    munmap(map->data, map->alloc);
	    if (ftruncate(map->fd, size) < 0) {
		fprintf(stderr, "resize_map: truncate failed\n");
		perror("");
		exit(1);
	    }
	    if (lseek(map->fd, 0, SEEK_SET) < 0)
	      error("set_map_size: seek failed");
	    map->data = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, map->fd, 0);
	    if (map->data < (void *) 0)
	      error("set_map_size: Remap failed");
	    map->alloc = size;
	}
    }
}

