#include <spa.h>


#define MAX_TOKEN 127

/* null 0 */
#define STRING 1
#define LBRACE 2
#define RBRACE 3
#define COLON 4
#define SEMICOLON 5
#define COMMA 6
#define VBAR 7
#define STAR 8
#define PLUS 9
#define QMARK 10
#define LPAREN 11
#define RPAREN 12
#define EQUALS 13
#define ARROW 14
#define ATSIGN 15
#define LBRACKET 16
#define RBRACKET 17
#define HASH 18
#define PERCENT 19
#define BANG 20
#define DOLLAR 21
#define AMPERSAND 22
#define NTOKTYPES 23


struct tokstream {
    FILE *file;
    char *filename;
    int line;
    char token[MAX_TOKEN+1];
    int la;
};


static char *TokTypeNames[] = {
  "NULL", "string", "lbrace", "rbrace", "colon", "semicolon", "comma", "vbar",
  "star", "plus", "qmark", "lparen", "rparen", "equals", "arrow", "atsign",
  "lbracket", "rbracket", "hash", "percent", "bang", "dollar", "ampersand"
};


char *toktype_string (int t)
{
    if (t == EOF) return "EOF";
    else if (t >= NTOKTYPES) return "(bad toktype)";
    else return TokTypeNames[t];
}


static char char_code[] = {
    /* |000 NUL|001 SOH|002 STX|003 ETX|004 EOT|005 ENQ|006 ACK|007 BEL| */
    0,      0,      0,      0,      0,      0,      0,      0,

    /* |010 BS |011 HT |012 NL |013 VT |014 NP |015 CR |016 SO |017 SI | */
    0,      0,      0,      0,      0,      0,      0,      0,

    /* |020 DLE|021 DC1|022 DC2|023 DC3|024 DC4|025 NAK|026 SYN|027 ETB| */
    0,      0,      0,      0,      0,      0,      0,      0,

    /* |030 CAN|031 EM |032 SUB|033 ESC|034 FS |035 GS |036 RS |037 US | */
    0,      0,      0,      0,      0,      0,      0,      0,
    
    /* |040 SP |041  ! |042  " |043  # |044  $ |045  % |046  & |047  ' | */
    0,      BANG,      0,    0,   DOLLAR, PERCENT,   AMPERSAND,      0,
    
    /* |050  ( |051  ) |052  * |053  + |054  , |055  - |056  . |057  / | */
    LPAREN, RPAREN,   STAR,   PLUS,  COMMA,      0,      0,      0,
    
    /* |060  0 |061  1 |062  2 |063  3 |064  4 |065  5 |066  6 |067  7 | */
    0,      0,      0,      0,      0,      0,      0,      0,
    
    /* |070  8 |071  9 |072  : |073  ; |074  < |075  = |076  > |077  ? | */
    0,      0,  COLON, SEMICOLON,   0, EQUALS,      0,  QMARK,
    
    /* |100  @ |101  A |102  B |103  C |104  D |105  E |106  F |107  G | */
    ATSIGN,      0,      0,      0,      0,      0,      0,     0,
    
    /* |110  H |111  I |112  J |113  K |114  L |115  M |116  N |117  O | */
    0, 0, 0, 0, 0, 0, 0, 0,
    
    /* |120  P |121  Q |122  R |123  S |124  T |125  U |126  V |127  W | */
    0, 0, 0, 0, 0, 0, 0, 0,
    
    /* |130  X |131  Y |132  Z |133  [ |134  \ |135  ] |136  ^ |137  _ | */
    0, 0, 0, LBRACKET, 0, RBRACKET, 0, 0,
    
    /* |140  ` |141  a |142  b |143  c |144  d |145  e |146  f |147  g | */
    0, 0, 0, 0, 0, 0, 0, 0,
    
    /* |150  h |151  i |152  j |153  k |154  l |155  m |156  n |157  o | */
    0, 0, 0, 0, 0, 0, 0, 0,
    
    /* |160  p |161  q |162  r |163  s |164  t |165  u |166  v |167  w | */
    0, 0, 0, 0, 0, 0, 0, 0,
    
    /* |170  x |171  y |172  z |173  { |174  | |175  } |176  ~ |177 DEL| */
    0, 0, 0, LBRACE, VBAR, RBRACE, 0, 0
};


int unscan_token (int type, struct tokstream *S)
{
    if (S->la) error("unscan_token: repeated unscan");
    S->la = type;
}


int peek_token (struct tokstream *S)
{
    if (!S->la) S->la = scan_token(S);
    return S->la;
}


int scan_token (struct tokstream *S)
{
    int t, c, i;
    FILE *file = S->file;

    if (S->la) {
	t = S->la;
	S->la = 0;
    }
    else {
	i = 0;
	S->token[0] = '\0';
	S->line += skip_whitespace_and_comments(file);
	c = read_char(file);
	if (c == EOF) return EOF;

	/** String in Quotes **/
	if (c == '"') {
	    while ((c = read_char(file)) != '\"') {
		if (c == EOF) error("scan_token: premature EOF");
		if (c == '\\') {
		    c = read_char(file);
		    if (c == EOF) error("scan_token: premature EOF");
		}
		if (i >= MAX_TOKEN) error("scan_token: buffer overflow");
		S->token[i++] = c;
		if (c == '\n') S->line++;
	    }
	    S->token[i] = '\0';
	    t = STRING;
	}

	/** Arrow **/
	else if (c == '-' && peek_char(file) == '>') {
	    read_char(file);
	    t = ARROW;
	}

	/** Single-char Special Symbol **/
	else if (c < 128 && char_code[c]) {
	    t = char_code[c];
	}

	/** Symbol **/
	else {
	    while (1) {
		t = STRING;
		if (i >= MAX_TOKEN) error("scan_token: buffer overflow");
		if (c == '\\') {
		    c = read_char(file);
		    if (c == EOF) error("scan_token: premature EOF");
		    if (c == '\n') S->line++;
		}
		S->token[i++] = c;
		c = read_char(file);
		if (c == EOF || isspace(c)) {
		    S->token[i] = '\0';
		    if (c == '\n') S->line++;
		    break;
		}
		else if (c < 128 && char_code[c] || c == '"') {
		    unread_char(c, file);
		    S->token[i] = '\0';
		    break;
		}
	    }
	}
    }
    return t;
}


int skip_whitespace_and_comments (FILE *file)
{
    int c, lines = 0;

    while (isspace(c = peek_char(file)) || c == '#') {
	read_char(file);
	if (c == '#') {
	    while ((c = read_char(file)) != '\n' && c != EOF) ;
	}
	if (c == '\n') lines++;
    }
    return lines;
}


struct tokstream *make_tokstream (FILE *file)
{
    struct tokstream *S;

    allocate(S, 1, struct tokstream);
    S->file = file;
    if (file == stdin) S->filename = "<stdin>";
    else S->filename = NULL;
    S->line = 1;
    return S;
}


struct tokstream *open_tokstream (char *filename)
{
    struct tokstream *S;

    allocate(S, 1, struct tokstream);
    if (filename == NULL) {
	S->file = stdin;
	S->filename = "<stdin>";
    }
    else {
	S->file = Fopen(filename, "r");
	S->filename = strdup(filename);
    }
    S->line = 1;
    return S;
}

    
int close_tokstream (struct tokstream *stream)
{
    if (stream->file != stdin) fclose(stream->file);
    if (stream->filename != NULL) free(stream->filename);
    free(stream);
}
