#include <stdio.h>
#include "spa.h"
#include "aarray.h"
#include "dict.h"


/**-------------------------------------------------------------------
***  String Array
**/

struct string_array {
  struct aarray *strings;
  struct aarray *chars;
};

struct string_array *make_string_array (int alloc)
{
  struct string_array *a;

  allocate(a, 1, struct string_array);
  a->strings = make_aarray(alloc, sizeof(int));
  a->chars = make_aarray(alloc * 6, sizeof(char));
  return a;
}

void init_string_array (struct string_array *a, int alloc)
{
  a->strings = make_aarray(alloc, sizeof(char *));
  a->chars = make_aarray(alloc * 6, sizeof(char));
}  
  
char *string_array_ref (struct string_array *a, int i)
{
  int offset;

  aget(offset, a->strings, i);
  return aref(a->chars, offset);
}

int string_array_insert (char *s, struct string_array *a)
{
  int i = a->strings->len;
  int k = a->chars->len;
  
  apush(k, a->strings);
  aallocate(a->chars, strlen(s) + 1);
  strcpy(a->chars->contents + k, s);
  return i;
}

#define string_array_len(a) ((a)->strings->len)

void write_string_array (struct string_array *a, FILE *file)
{
  write_aarray(a->strings, file);
  write_aarray(a->chars, file);
}

void read_string_array (struct string_array *a, FILE *file)
{
  a->strings = read_aarray(file);
  a->chars = read_aarray(file);
}

zero_string_array (struct string_array *a)
{
    azero(a->strings);
    azero(a->chars);
}



/**-------------------------------------------------------------------
***  String Table
**/

struct string_table {
  struct dict *d;
  struct string_array a;
};

#define string_table_len(tbl) ((tbl)->d->count)

struct string_table *scan_string_table (int alloc, FILE *file)
{
  int i;
  struct record r;
  struct string_table *tbl;

  allocate(tbl, 1, struct string_table);
  tbl->d = make_dict(alloc);
  init_string_array(&tbl->a, alloc);

  while (scan_record(&r, file) != EOF) {
    if (r.nfields != 2) {
      fprintf(stderr, "Warning: scan_string_table: bad record: ");
      print_record(&r, stderr);
    }
    i = pat_insert(r.fields[0], tbl->d);
    if (!new_entry_p)
      error("Multiply specified entry in string table for %s", r.fields[0]);
    if (string_array_insert(r.fields[1], &tbl->a) != i)
      error("scan_string_table: arrays out of synch");
  }
  return tbl;
}

char *string_table_ref (char *s, struct string_table *tbl)
{
  int i = pat_search(s, tbl->d);
  if (i < 0) return NULL;
  return string_array_ref(&tbl->a, i);
}

#define string_table_key(i, tbl) (pat_key((i), (tbl)->d))
#define string_table_value(i, tbl) (string_array_ref(&(tbl)->a, (i)))

void dump_string_table (struct string_table *tbl, FILE *file)
{
  int i;

  for (i = 0; i < string_table_len(tbl); i++) {
    fprintf(file, "%s\t%s\n",
	    string_table_key(i, tbl),
	    string_table_value(i, tbl));
  }
}

void write_string_table (struct string_table *tbl, FILE *file)
{
  write_dict(tbl->d, file);
  write_string_array(&tbl->a, file);
}

struct string_table *read_string_table (FILE *file)
{
  struct string_table *tbl;

  allocate(tbl, 1, struct string_table);
  tbl->d = read_dict(file);
  read_string_array(&tbl->a, file);
  return tbl;
}

void save_string_table (struct string_table *tbl, char *filename)
{
  FILE *file = Fopen(filename, "w");
  write_string_table(tbl, file);
  fclose(file);
}

struct string_table *load_string_table (char *filename)
{
  FILE *file = Fopen(filename, "r");
  struct string_table *tbl = read_string_table(file);
  fclose(file);
  return tbl;
}


int compile_string_table_cmd (int argc, char *argv[])
{
    char *s = NULL, filename[MAXFILENAME];
    FILE *file;
    struct string_table *tbl;
    
    parse_args(&argc, &argv, "s", &s);
    if (s == NULL) error("No filename provided");
    strcpy(filename, s);

    file = Fopen(filename, "r");
    tbl = scan_string_table(1000, file);
    fclose(file);

    chsuf(filename, ".txt", ".stb");
    save_string_table(tbl, filename);
    return 0;
}

int dump_string_table_cmd (int argc, char *argv[])
{
  char *path;
  struct string_table *tbl;
  
  parse_args(argc, argv, "s", &path);
  tbl = load_string_table(path);
  dump_string_table(tbl, stdout);
  return 0;
}
