#include <stdio.h>
#include "spa.h"
#include "aarray.h"


/**
***  What makes this slightly nontrivial is that we don't want to
***  insist that the selected line numbers be sorted, yet we also
***  don't want to read the text file repeatedly.
***
***  We read the line numbers into an array whose entries are:
***
***      input-seq output-seq line
***
***  where input-seq is the line number (initially unsorted) and
***  output-seq is the sequence in the sample (initially sorted).
***
***  Then we sort this array by output-seq.
***
***  Then we read once through the text file, storing the selected
***  lines in this array.
***
***  Then we sort the array again by output-seq, and pass through
***  outputing the lines in the order originally specified in the
***  lines file.
**/

struct entry {
    int txt;
    int smp;
    char *line;
};

int SelectAlloc = 500;

struct aarray *LinesA;
struct entry *Lines;

#define NLines (LinesA->len)

static int smpcmp (e1, e2)
    struct entry *e1, *e2;
{
    return e1->smp - e2->smp;
}

static int txtcmp (e1, e2)
    struct entry *e1, *e2;
{
    return e1->txt - e2->txt;
}


#define LINESFILE  1
#define LINESLIST  2
#define LINESRANGE 3


#define new_line(line) \
  (e = apull(LinesA), \
   e->smp = current++, \
   e->txt = (line))


select_lines (int type, void *lines, int n, FILE *infile, FILE *outfile)
{
    char line[1024];
    int i, current = 0, target, *lineNumbers;
    struct entry *e;

    LinesA = make_aarray(SelectAlloc, sizeof(struct entry));

    /**  Scan in the lines  **/
    switch (type) {

      case LINESFILE:
	while (scan_int(target, lines) != EOF) {
	    new_line(target);
	}
	break;

      case LINESLIST:
	  lineNumbers = (int *) lines;
	  for (i = 0; i < n; ++i) {
	      target = lineNumbers[i];
	      new_line(target);
	  }
	  break;

      case LINESRANGE:
	for (target = ((int *)lines)[0]; target <= ((int *)lines)[1]; target++) {
	    new_line(target);
	}
	break;

      default: error("select_lines: Bad type");
    }

    Lines = (struct entry *) LinesA->contents;

    /**  Sort by txt sequence  **/
    qsort(Lines, NLines, sizeof(struct entry), txtcmp);

    /**  Extract lines  **/
    current = -1;
    for (i = 0; i < NLines; i++) {
	target = Lines[i].txt;
	if (target < current) error("sort failed");
	if (current < target) {  /* careful: it's possible to select the same line twice */
	    while (++current < target) {
		skip_line(infile);
	    }
	    if (scan_line(line, 1024, infile) == EOF) {
		fprintf(stderr, "Premature end of text file");
		exit(1);
	    }
	}
	Lines[i].line = copy_string(line);
    }

    /**  Resort by smp sequence  **/
    qsort(Lines, NLines, sizeof(struct entry), smpcmp);

    /**  Output  **/
    for (i = 0; i < NLines; i++) {
	fprintf(outfile, "%s\n", Lines[i].line);
    }
}


main (int argc, char *argv[])
{
    char *infilename = NULL, *linesfilename = NULL, *outfilename = NULL, *lines = NULL, *tmp;
    FILE *infile = NULL, *linesfile = NULL, *outfile = stdout;
    int a[64], n;

    parse_args(&argc, &argv, "?ss-sf-so", &infilename, &lines, &linesfilename, &outfilename);

    if (infilename != NULL && is_number(infilename) && (lines == NULL || !is_number(lines)) ||
	lines != NULL && !is_number(lines) && infilename != NULL && is_number(infilename)) {
	tmp = lines;
	lines = infilename;
	infilename = tmp;
    }

    if (infilename != NULL) infile = Fopen(infilename, "r");
    if (linesfilename != NULL) linesfile = Fopen(linesfilename, "r");
    if (outfilename != NULL) outfile = Fopen(outfilename, "w");
    
    if (infile == NULL) {
	infile = stdin;
	if (linesfile == NULL && lines == NULL) error("No lines specified");
    }
    else if (linesfile == NULL && lines == NULL) linesfile = stdin;

    if (lines == NULL) {
	select_lines(LINESFILE, linesfile, 0, infile, outfile);
    }
    else if ((tmp = string_member('-', lines)) != NULL) {
	a[0] = atoi(lines);
	a[1] = atoi(tmp+1);
	select_lines(LINESRANGE, a, 2, infile, outfile);
    }
    else {
	n = atois(lines, a, 64);
	select_lines(LINESLIST, a, n, infile, outfile);
    }

    exit(0);
}
