#include <stdio.h>
#include <sys/time.h>
#include <limits.h>
#include <spa.h>
#include <spamap.h>
#include "sents.h"


/**
***  This is set up to read corpora now.  There are two kinds of corpus
***  that we want to be able to deal with right now.
***
***  The minimal corpus is created by calling sents -c on a .tag file.
***  This creates a token file indexed by corpus position and a sentence
***  file giving the start and end positions of each sentence.  These
***  indices point into the original text file; no separate lexicon is
***  set up.
***
***  Alternately, we use xkwic corpora.  They involve a separate lexicon.
***  
***  Both types of corpus use (some of) the following arrays (which are
***  stored in files).
***
***      C     Corpus.  C[t] is the word at the corpus-position t.
***            t starts at 0.
***
***      I     Lexicon Index.  I[w] is the string-offset for the word w.
***            w starts at 0.
***
***      L     Lexicon.  L+i is the string-address for the string-offset i.
***            i starts at 0.
***
***      S     Sentences.  S[s][0] is the corpus-position of the first word
***            in the sentence s, and S[s][1] is the last word (inclusive).
***            s starts at 0.
***
***  The differences between a minimal corpus and an xkwic corpus are the
***  following:
***
***      In a minimal corpus, I does not exist.  Words are identical to
***      string-offsets.  In an xkwic corpus, the string for corpus-position
***      t starts at L+I[C[t]] and is '\0'-terminated.  In a minimal corpus,
***      it starts at L+C[t] and is whitespace-terminated.
***
***      A minimal corpus is not registered.  We just call find_file on
***      the name of the text file, rather than going through xkwic_corpus_directory.
***      The filenames are also different:
***
***                Xwic                     minimal
***
***          C    word.corpus             <name>.tdx
***          I    word.lexicon.idx        --
***          L    word.lexicon            <name>.tag
***          S    s.rng                   <name>.snt
***
***          P    pos.corpus
***          J    pos.lexicon.idx
***          M    pos.lexicon
***
***  We localize these differences in print_token(), copy_token(), and
***  open_corpus().
**/



/** Formatting: **/

int PrintRaw = 0;
int TraceSentSearch = 0;
int LeftContext = DEFAULT_LEFT_CONTEXT;
int LabelLen = 0;
int GutterWidth = DEFAULT_GUTTER_WIDTH;
char *WrapMarker = "//";

char *Bos = "<s>";
char *Eos = "</s>";


/** Corpus: **/

int CorpusType = 0;
char *CorpusName;
char *CorpusDir;
struct map *WordCorpusMap;
struct map *WordLexiconIdxMap;
struct map *WordLexiconMap;
struct map *PosCorpusMap = NULL;
struct map *PosLexiconIdxMap;
struct map *PosLexiconMap;
struct map *SRngMap;
int *C;
int *I;
char *L;
int *P;
int *J;
char *M;
int (*S)[2];
int NSents;
int NTokens;
int NWordTypes;

#define XKWIC   1
#define MINCORP 2

#define open_corpus(name) \
  (open_minimal_corpus(name) || open_xkwic_corpus(name) || error("Can't find corpus %s", (name)))

#define describe_corpus(name, file) \
  (CorpusType == XKWIC ? describe_xkwic_corpus(name, file) : describe_minimal_corpus(name, file))

#define print_token(t, file) \
  (CorpusType == XKWIC ? \
   print_string(L+I[C[t]], (file)) : \
   print_minimal_token((t), (file)))

#define print_tagged_token(t, file) \
  (CorpusType == XKWIC ? \
   fprintf((file), "%s\t%s\n", L+I[C[t]], M+J[P[t]]) : \
   print_min_tagged_token((t), (file)))

#define copy_token(t, s) \
  (CorpusType == XKWIC ? \
   (strcpy((s), L+I[C[t]]), strlen(L+I[C[t]])) : \
   copy_minimal_token((t), (s)))





/**  Common to both styles  **/

close_corpus ()
{
    close_map(WordCorpusMap);
    if (WordLexiconIdxMap != NULL) close_map(WordLexiconIdxMap);
    close_map(WordLexiconMap);
    if (PosCorpusMap != NULL) {
	close_map(PosCorpusMap);
	close_map(PosLexiconMap);
	close_map(PosLexiconIdxMap);
    }
    close_map(SRngMap);
}


/**===================================================================
***
***			     Xkwic Corpus
**/

/**
***  xkwic_corpus_directory (name)
***
***  Name is e.g. "tb0p".  The return value is the directory in which the
***  corpus lives, e.g. "/home/statling/disk1/corpora/tb0p".  The returned
***  string is static and gets re-used each time xkwic_corpus_directory is called.
**/

char *xkwic_corpus_directory (char *name)
{
    static char buf[128];
    char *registry, filename[128];
    FILE *file;
    struct record r;

    registry = getenv("CORPUS_REGISTRY");
    if (registry == NULL) error("xkwic_corpus_directory: CORPUS_REGISTRY is not set");
    sprintf(filename, "%s/%s", registry, name);
    file = Fopen(filename, "r");
    
    while (scan_record_1(&r, file, " \t", SQUEEZE) != EOF) {
	if (r.nfields >= 2 && string_equal(r.fields[0], "HOME")) {
	    strcpy(buf, r.fields[1]);
	    fclose(file);
	    return buf;
	}
    }
    return NULL;
}

/**
***  open_xkwic_corpus (name)
***
***  Opens the four files that we're interested in, setting the variables
***  WordCorpus, WordLexiconIdx, WordLexicon, and SRng.
**/

int open_xkwic_corpus (char *name)
{
    char *dir, filename[128];

    dir = xkwic_corpus_directory(name);
    if (dir == NULL) return 0;

    CorpusType = XKWIC;

    sprintf(filename, "%s/word.corpus", dir);
    WordCorpusMap = open_map(filename, "r");
    C = WordCorpusMap->data;
    NTokens = WordCorpusMap->alloc / sizeof(int);

    sprintf(filename, "%s/word.lexicon.idx", dir);
    WordLexiconIdxMap = open_map(filename, "r");
    I = WordLexiconIdxMap->data;
    NWordTypes = WordLexiconIdxMap->alloc / sizeof(int);

    sprintf(filename, "%s/word.lexicon", dir);
    WordLexiconMap = open_map(filename, "r");
    L = WordLexiconMap->data;

    sprintf(filename, "%s/pos.corpus", dir);
    PosCorpusMap = open_map(filename, "r");
    P = PosCorpusMap->data;

    sprintf(filename, "%s/pos.lexicon.idx", dir);
    PosLexiconIdxMap = open_map(filename, "r");
    J = PosLexiconIdxMap->data;

    sprintf(filename, "%s/pos.lexicon", dir);
    PosLexiconMap = open_map(filename, "r");
    M = PosLexiconMap->data;

    sprintf(filename, "%s/s.rng", dir);
    SRngMap = open_map(filename, "r");
    S = SRngMap->data;
    NSents = SRngMap->alloc / (2*sizeof(int));

    return 1;
}


describe_xkwic_corpus (char *name, FILE *file)
{
    fprintf(file, "Corpus:     %s/%s\n", xkwic_corpus_directory(name), name);
    fprintf(file, "Tokens:     %d\n", NTokens);
    fprintf(file, "Word Types: %d\n", NWordTypes);
    fprintf(file, "Sents:      %d\n", NSents);
}




/**===================================================================
***
***			    Minimal Corpus
**/


/**
***  open_minimal_corpus (filename)
**/

int open_minimal_corpus (char *name)
{
    char filename[MAXFILENAME];
    FILE *file;

    strcpy(filename, name);
    addsuf(filename, ".tag");
    if (find_datafile(filename) == NULL) return 0;
    
    CorpusType = MINCORP;
    
    WordLexiconMap = open_map(filename, "r");
    L = WordLexiconMap->data;

    chsuf(filename, ".tag", ".tdx");
    WordCorpusMap = open_map(filename, "r");
    C = WordCorpusMap->data;
    NTokens = WordCorpusMap->alloc / sizeof(int);

    WordLexiconIdxMap = NULL;

    chsuf(filename, ".tdx", ".snt");
    SRngMap = open_map(filename, "r");
    S = SRngMap->data;
    NSents = SRngMap->alloc / (2*sizeof(int));

    return 1;
}

describe_minimal_corpus (char *name, FILE *file)
{
    char filename[MAXFILENAME];

    strcpy(filename, name);
    addsuf(filename, ".tag");
    if (find_datafile(filename) == NULL)
      error("describe_minimal_corpus: %s not found", filename);
    
    fprintf(file, "Textfile:   %s\n", filename);

    chsuf(filename, ".tag", ".tdx");
    fprintf(file, "Tokenfile:  %s\n", filename);

    chsuf(filename, ".tdx", ".snt");
    fprintf(file, "Sentfile:   %s\n", filename);

    fprintf(file, "Tokens:     %d\n", NTokens);
    fprintf(file, "Sents:      %d\n", NSents);
}


/**
***  index_sents (textfile, tokenfile, sentfile)
***
***  Scans textfile for <s> and </s> markup and creates a token file and
***  a sent file.
***
***  The token file is like word.corpus in that it has one int for each
***  token.  However, unlike word.corpus, the ints are offsets of lines
***  in the original text file, not word ids.
***
***  The sent file has the same format as an s.rng file.  It contains
***  pairs of ints representing the closed interval [i,j] of corpus
***  positions contained in the sentence.
**/


index_sents (char *filename)
{
    char buf[MAXFILENAME];
    FILE *textfile, *tokenfile, *sentfile;

    CorpusType = MINCORP;

    strcpy(buf, filename);
    filename = buf;

    addsuf(filename, ".tag");
    textfile = Fopen(filename, "r");

    chsuf(filename, ".tag", ".tdx");
    tokenfile = Fopen(filename, "w");

    chsuf(filename, ".tdx", ".snt");
    sentfile = Fopen(filename, "w");

    index_sents_1(textfile, tokenfile, sentfile);

    fclose(textfile);
    fclose(tokenfile);
    fclose(sentfile);
}


#define write_sent_entry() \
  (bos < t ? (write_int(bos, sentfile), write_int(t-1, sentfile)) : \
   fprintf(stderr, "Warning: empty sentence at %d, skipping\n", ct))

index_sents_1 (FILE *textfile, FILE *tokenfile, FILE *sentfile)
{
    char line[1024];
    int ct = 0, t = 0, bos = -1;

    while (scan_line(line, 1024, textfile) != EOF) {
	if (string_member('\t', line) == NULL) {
	    if (bos >= 0) {
		write_sent_entry();
		bos = -1;
	    }
	}
	else {
	    if (bos < 0) bos = t;
	    t++;
	    write_int(ct, tokenfile);
	}
	ct += strlen(line) + 1;
    }
    if (bos >= 0) write_sent_entry();
}
	  


/**
***  print_minimal_token (t, file)
**/

print_minimal_token (int t, FILE *file)
{
    char *s = L+C[t];

    while (!isspace(*s)) print_char(*s++, file);
}

print_min_tagged_token (int t, FILE *file)
{
    char *s = L+C[t];

    while (*s != '\n') print_char(*s++, file);
    print_char('\n', file);
}

copy_minimal_token (int t, char *s)
{
    int len = 0;
    char *u = L+C[t];

    while (!isspace(*u)) *s++ = *u++, len++;
    *s = '\0';
    return len;
}


/**===================================================================
***
***		     Print Sents, Word in Context
**/


/**
***  print_sent (s, file)
***
***  Print the s-th sentence.
**/

print_sent (int s, FILE *file)
{
    int start, end, t;

    if (s < 0 || s >= NSents) error("print_sent: %d out of range", s);

    start = S[s][0];
    end = S[s][1];

    if (PrintRaw) {
	print_string(Bos, file);
	terpri(file);
	for (t = start; t <= end; t++) print_tagged_token(t, file);
	print_string(Eos, file);
	terpri(file);
    }
    else {
	fprintf(file, "%d:", s);
	for (t = start; t <= end; t++) {
	    print_char(' ', file);
	    print_token(t, file);
	}
	terpri(file);
    }
}


/**
***  find_token_sent (t)
***
***  Returns the index of the sentence containing the given token.
***
***  [u,v] is an inclusive range of sentence indices to be searched.
***  We're looking for a sentence x = [x0,x1] such that x0 <= t <= x1.
***  We do an interpolation search.  We start with the ranges [0,NSents-1]
***  and [0,NTokens-1].
***
***  Given ranges [u,v] and [i,j], define:
***
***      n = # of tokens in interval = j-i+1
***      N = # of sentences in interval = v-u+1
***      k = ave length of sentence in interval = n/N
***      dt = # of tokens in interval before t = t-i
***      du = expected # of sents in interval before target = dt/k
***
***  Our best guess about the target is x = u+[du], where [] indicates
***  the integral part.  If we declare n, N, and dt to be ints, then setting
***  Du = dt*N/n is equal to [du] because of truncation in integer division.
***
***  Given x = u+Du, let [x0,x1] be the token interval of sentence x.  We have
***  three cases:
***
***      x0 <= t <= x1         Success
***      t < x0                Search [u,x-1] and [i,x0-1]
***      x1 < t                Search [x+1,v] and [x1+1,j]
**/

int find_token_sent (int t, int *sent, int *start, int *end)
{
    int u, v, x, i, j, x0, x1;

    u = i = 0;
    v = NSents-1;
    j = NTokens-1;

    if (TraceSentSearch) fprintf(stderr, "Searching for sent of token %d\n", t);

    while (1) {
	x = u + (t-i)*(v-u+1)/(j-i+1);
	x0 = S[x][0];
	x1 = S[x][1];

	if (TraceSentSearch)
	  fprintf(stderr, "  Searching s-interval [%d,%d] t-interval [%d,%d], x=%d=[%d,%d]\n",
		  u, v, i, j, x, x0, x1);

	if (x0 <= t && t <= x1) {
	    *sent = x;
	    *start = x0;
	    *end = x1;
	    return 1;
	}
	if (t < x0) {
	    v = x-1;
	    j = x0-1;
	}
	else {
	    u = x+1;
	    i = x1+1;
	}
    }
}



kwic_init ()
{
    int x = 100, sidlen, tidlen;

    sidlen = 2;
    while (x <= NSents) {
	x *= 10;
	sidlen++;
    }

    x = 1000;
    tidlen = 3;
    while (x <= NTokens) {
	x *= 10;
	tidlen++;
    }

    LabelLen = sidlen + tidlen + 3;
    LeftContext -= LabelLen;
}


/**
***  print_word_in_context (t, file)
***
***  Print the t-th token in the context of its sentence.  There is a fixed-
***  length LeftContext and a fixed GutterWidth.  Left context that exceeds the
***  fixed window gets wrapped around, and is printed following a WrapMarker.
**/

print_word_in_context (int t, FILE *file)
{
    int sent, i, j, k, len, buflen = 0, start;
    char *s, buf[1024], idbuf[32];

    if (t < 0 || t > NTokens) error("print_word_in_context: %d out of range", t);

    find_token_sent(t, &sent, &i, &j);
    if (LabelLen == 0) kwic_init();

    for (k = i; k < t; k++) {
	buflen += copy_token(k, buf+buflen);
	buf[buflen++] = ' ';
	if (buflen > 1024) error("print_word_in_context: buffer overflow");
    }
    if (buflen > 0) buflen--;  /* the last space */
    buf[buflen] = '\0';
    start = buflen <= LeftContext ? 0 : buflen-LeftContext;

    sprintf(idbuf, "%d/%d: ", sent, t);
    print_field(LabelLen, idbuf, file);
    print_field(-LeftContext, buf+start, file);
    buf[start] = '\0';
    for (k = 0; k < GutterWidth-1; k++) print_char(' ', file);
    for (k = t; k <= j; k++) {
	print_char(' ', file);
	print_token(k, file);
    }
    if (start > 0) {
	print_char(' ', file);
	print_string(WrapMarker, file);
	print_char(' ', file);
	print_string(buf, file);
    }
    terpri(file);
}



/**-------------------------------------------------------------------
***  Toplevel
**/


int isrange (char *s)
{
    if (s == NULL) return 0;
    for (; *s; s++) {
	if (!(isdigit(*s) || *s == ',' || *s == '-' || isspace(*s)))
	  return 0;
    }
    return 1;
}


/**
***  print_sents (u, v, file)
***
***  Print sentences in range [u,v] inclusive.
**/

int print_sents (char *range, FILE *file)
{
    char *second;

    if (!isrange(range)) {
	int t;
	FILE *infile = Fopen(range, "r");
	while (scan_int(t, infile) != EOF) {
	    print_sent(t, file);
	}
	fclose(infile);
    }

    else if ((second = string_member('-', range)) != NULL) {
	int s, i, j;
	i = atoi(range);
	j = atoi(second+1);
	for (s = i; s <= j; s++) {
	    if (s >= NSents) break;
	    print_sent(s, file);
	}
    }

    else {
	int a[64], n, i;
	n = atois(range, a, 64);
	for (i = 0; i < n; i++) {
	    print_sent(a[i], file);
	}
    }
}



/**
***  print_kwic
**/

print_kwic (char *tokens, FILE *outfile)
{
    if (isrange(tokens)) {
	int a[64], i, n;
	n = atois(tokens, a, 64);
	for (i = 0; i < n; i++) {
	    print_word_in_context(a[i], outfile);
	}
    }

    else {
	int t;
	FILE *infile = Fopen(tokens, "r");
	while (scan_int(t, infile) != EOF) {
	    print_word_in_context(t, outfile);
	}
	fclose(infile);
    }
}



/**===================================================================
***
***				 Main
**/


int main (int argc, char *argv[])
{
    char *arg1 = NULL, *arg2 = NULL;
    char *outfilename = NULL, *name = NULL, *range = NULL, *rangefile = NULL, *tmp;
    int compile = 0, describe = 0, tokens = 0;
    FILE *outfile = stdout;

    Usage = "-cdtr range [[+]infile] [-o outfile]\n\
\n\
Options:\n\
-c   compile: create minimal corpus from tag file\n\
-d   describe corpus\n\
-t   range is tokens, not sents; print kwic\n\
-T   tagged output\n\
-o   file to receive output\n\
+    prefixed to infile indicates an xkwic corpus\n\
\n\
Defaults:\n\
KwicLeftContext    left context in kwic\n\
KwicGutterWidth    gutter width in kwic\n\
KwicWrapMarker     string to insert at wrap point\n\
Bos                beginning of sentence marker (raw mode)\n\
Eos                end of sentence marker (raw mode)";

    defaulti(TraceSentSearch, "TraceSentSearch");
    defaulti(LeftContext, "KwicLeftContext");
    defaulti(GutterWidth, "KwicGutterWidth");
    defaults(WrapMarker, "KwicWrapMarker");
    defaults(Bos, "Bos");
    defaults(Eos, "Eos");

    parse_args(&argc, &argv, "?ss-bc-so-bd-bt-bvsent-bT-sf",
	       &arg1, &arg2, &compile, &outfilename,
	       &describe, &tokens, &TraceSentSearch, &PrintRaw, &rangefile);

    /** with -c, the first arg is the filename **/
    if (compile || describe) {
	if (arg2 != NULL) error("usage: %s", Usage);
	name = arg1;
    }
    else {
	range = arg1;
	name = arg2;
    }
    
    if (rangefile != NULL) {
	if (range) error("Don't supply both range and -f");
	range = rangefile;
    }

    /**  Compile  **/
    if (compile) {
	index_sents(name);
    }

    /**  Print  **/
    else {
	if (name == NULL) {
	    defaults(name, "CORPUS");
	    if (name == NULL) error("No corpus");
	}

	if (outfilename != NULL) outfile = Fopen(outfilename, "w");

	open_corpus(name);
	if (describe) describe_corpus(name, outfile);
	else if (tokens) print_kwic(range, outfile);
	else print_sents(range, outfile);
	close_corpus();
    }

    exit(0);
}
