#include <spa.h>
#include <amgr.h>
#include <regdict.h>
#include <cass.h>


int Pairs = 0;


int prin_tuple (struct phrase *p, int props, FILE *file);


main (int argc, char *argv[])
{
    char *filename = NULL, *grammar = NULL, *tuple_specs = NULL;
    FILE *file = stdin;

    parse_args(&argc, &argv, "?s-sg-sf-bp", &filename, &grammar, &tuple_specs, &Pairs);

    load_cass_grammar(grammar);
    load_tuple_specs(tuple_specs);
 
    if (filename != NULL) file = Fopen(filename, "r");
    while (scan_sentence(file) != EOF) {
 	parse_sentence();
	prin_tuples(stdout);
    }
    return 0;
}

	


					  


/**************************************************************************/
/**                                                                      **/
/**  Tuple Specs                                                         **/
/**                                                                      **/
/**               src0 (s,h=vx)                                          **/
/**               orc0 (s,h=vx)                                          **/
/**               c0 (s,h=vx)                                            **/
/**               subc0 (s,h=vx)                                         **/
/**                                                                      **/
/**               c-inv (h=vx,s)                                         **/
/**                                                                      **/
/**               c (h=c0,o)                                             **/
/**               subc (h=subc0,o)                                       **/
/**               rc (h=src0/orc0,o)                                     **/
/**               vnp (h=vnx,o)                                          **/
/**               vgp (h=vgx,o)                                          **/
/**               infp (h=inf,o)                                         **/
/**               vp (h=vx,o)                                            **/
/**                                                                      **/
/**  root of verbal projection: c-inv, c, subc, rc, vnp, vgp, infp, vp   **/
/**  complex heads: c0, subc0, src0, orc0                                **/
/**                                                                      **/
/**  has subject: src0, orc0, c0, subc0, c-inv                           **/
/**                                                                      **/
/**  has object and compls: c, subc, rc, vnp, vgp, infp, vp              **/
/**                                                                      **/
/**  We walk through the tree looking for roots of verbal projections.   **/
/**  When we find one, we emit a tuple.  To emit a tuple, we find the    **/
/**  head string (chase head links), then we emit args associated with   **/
/**  the current node, then if the head is complex, we emit args         **/
/**  associated with the head.                                           **/
/**                                                                      **/
/**************************************************************************/


struct amgr *PropertyMgr = NULL;
int *Properties;
struct dict *LightDict = NULL;

AttId HeadRole = 0;
AttId SubjRole = 0;
AttId ObjRole = 0;
AttId FwRole = 0;
AttId PartRole = 0;

#define IS_VROOT          0x1
#define HAS_COMPLEX_HEAD  0x2
#define HAS_SUBJECT       0x4
#define HAS_COMPLS        0x8
#define IS_PP            0x10
#define IS_PREP          0x20
#define IS_PASSIVE       0x40
#define IS_BE            0x80
#define IS_IGNORE       0x100
#define IS_NP           0x200

#define HEAD_ROLE        -1
#define SUBJ_ROLE        -2
#define OBJ_ROLE         -3
#define FW_ROLE          -4
#define PART_ROLE        -5
#define IS_LIGHT         -6


load_tuple_specs (char *filename)
{
    FILE *file;
    char *fields[64], word[128];
    int n, sym, mask = 0, *p;
    
    if (filename == NULL) {
	defaults(filename, "TUPLE_SPECS");
	if (filename == NULL) filename = DEFAULT_TUPLE_SPECS;
    }
    file = Fopen(filename, "r");

    if (PropertyMgr == NULL) {
	PropertyMgr = make_amgr(&Properties, 100, sizeof(int));
	LightDict = make_dict(50);
    }

    while (skip_whitespace(file), scan_string(word, 128, " \t\n", file) != EOF) {
	if (string_equal(word, "v-root:")) mask = IS_VROOT;
	else if (string_equal(word, "complex-head:")) mask = HAS_COMPLEX_HEAD;
	else if (string_equal(word, "has-subject:")) mask = HAS_SUBJECT;
	else if (string_equal(word, "has-compls:")) mask = HAS_COMPLS;
	else if (string_equal(word, "pp:")) mask = IS_PP;
	else if (string_equal(word, "prep:")) mask = IS_PREP;
	else if (string_equal(word, "be:")) mask = IS_BE;
	else if (string_equal(word, "passive:")) mask = IS_PASSIVE;
	else if (string_equal(word, "ignore:")) mask = IS_IGNORE;
	else if (string_equal(word, "light:")) mask = IS_LIGHT;
	else if (string_equal(word, "np:")) mask = IS_NP;
	else if (string_equal(word, "head:")) mask = HEAD_ROLE;
	else if (string_equal(word, "subj:")) mask = SUBJ_ROLE;
	else if (string_equal(word, "obj:")) mask = OBJ_ROLE;
	else if (string_equal(word, "fw:")) mask = FW_ROLE;
	else if (string_equal(word, "part:")) mask = PART_ROLE;
	else if (mask == HEAD_ROLE) {
	    if (HeadRole) error("load_tuple_specs: multiple head roles");
	    HeadRole = intern_att(word);
	}
	else if (mask == SUBJ_ROLE) {
	    if (SubjRole) error("load_tuple_specs: multiple subj roles");
	    SubjRole = intern_att(word);
	}
	else if (mask == OBJ_ROLE) {
	    if (ObjRole) error("load_tuple_specs: multiple obj roles");
	    ObjRole = intern_att(word);
	}
	else if (mask == FW_ROLE) {
	    if (FwRole) error("load_tuple_specs: multiple fw roles");
	    FwRole = intern_att(word);
	}
	else if (mask == PART_ROLE) {
	    if (PartRole) error("load_tuple_specs: multiple part roles");
	    PartRole = intern_att(word);
	}
	else if (mask == IS_LIGHT) {
	    pat_insert(word, LightDict);
	}
	else {
	    if (word[strlen(word)-1] == ':') {
		error("load_tuple_specs: unrecognized property %s", word);
	    }
	    sym = intern(word);
	    gneed(PropertyMgr, sym);
	    SET(Properties[sym], mask);
	}
    }

    if (!(HeadRole && SubjRole && ObjRole && FwRole)) {
	error("Missing role(s):%s%s%s%s",
	      HeadRole ? "" : " head",
	      SubjRole ? "" : " subj",
	      ObjRole ? "" : " obj",
	      FwRole ? "" : " fw");
    }
}


/* This is obsolete, broken */

print_tuple_specs (FILE *file)
{
    int sym, mask;

    if (HeadRole) fprintf(file, "head: %s\n", att_string(HeadRole));
    if (SubjRole) fprintf(file, "subj: %s\n", att_string(SubjRole));
    if (ObjRole) fprintf(file, "obj: %s\n", att_string(ObjRole));
    for (sym = 1; sym <= glength(PropertyMgr); sym++) {
	mask = Properties[sym];
	if (mask) {
	    fprintf(file, "%s\t", symbol_string(sym));
	    if (GET(mask, IS_VROOT)) fprintf(file, "vroot ");
	    if (GET(mask, HAS_COMPLEX_HEAD)) fprintf(file, "complex-head ");
	    if (GET(mask, HAS_SUBJECT)) fprintf(file, "has-subject ");
	    if (GET(mask, HAS_COMPLS)) fprintf(file, "has-compls ");
	    if (GET(mask, IS_PP)) fprintf(file, "pp");
	    if (GET(mask, IS_PREP)) fprintf(file, "prep");
	    if (GET(mask, IS_PASSIVE)) fprintf(file, "passive");
	    terpri(file);
	}
    }
}


/**************************************************************************/
/**                                                                      **/
/**  Support Functions                                                   **/
/**                                                                      **/
/**************************************************************************/


inline int phrase_props (struct phrase *p)
{
    return p->cat <= glength(PropertyMgr) ? Properties[p->cat] : 0;
}


#define get_headword(hp,hsp,fp,p) (get_headword_1(hp,hsp,fp,NULL,p))

struct phrase *get_headword_1 (char *prefix, char **sp, struct phrase **fp, struct phrase **kp,
			       struct phrase *p)
{
    struct phrase *n, *f, *k;

    if (fp != NULL) *fp = NULL;
    if (kp != NULL) *kp = NULL;
    while (1) {
	if (fp != NULL) {
	    f = phrase_child(p, FwRole);
	    if (f != NULL) *fp = f;
	}
	if (kp != NULL) {
	    k = phrase_child(p, PartRole);
	    if (k != NULL) *kp = k;
	}
	n = phrase_child(p, HeadRole);
	if (n == NULL) break;
	p = n;
    }
    if (p->children == NULL) {
	*sp = word_string(p->i);
	prefix[0] = 0;
    }
    else {
	*sp = symbol_string(p->cat);
	prefix[0] = '%';
    }
    prefix[1] = 0;
    return p;
}
    

struct phrase *get_arg_head (char *hp, char **hsp, struct phrase **fp, struct phrase *arg)
{
    struct phrase *h, *k;

    h = get_headword_1(hp, hsp, fp, &k, arg);
    if (k != NULL && pat_search(*hsp, LightDict) != -1) {
	h = get_headword(hp, hsp, NULL, k);
    }
    return h;
}


char *compl_role (struct phrase *a, struct phrase *f)
{
    int props;
    
    props = phrase_props(a);
    if (GET(props, IS_PP)) {
	if (f == NULL) return symbol_string(a->cat);
	else if (f->children != NULL) return symbol_string(f->cat);
	else return word_string(f->i);
    }
    else if (GET(props, IS_PREP)) {
	if (a->children != NULL) return symbol_string(a->cat);
	else return word_string(a->i);
    }
    else return symbol_string(a->cat);
}


int is_passive (struct phrase *h, struct phrase *f)
{
    return h != NULL
        && f != NULL
        && GET(phrase_props(h), IS_PASSIVE)
        && GET(phrase_props(f), IS_BE)
	? 1 : 0;
}



/**************************************************************************/
/**                                                                      **/
/**  Print Tuples                                                        **/
/**                                                                      **/
/**************************************************************************/


prin_tuples (FILE *file)
{
    int i;
    
    for (i = 1; i <= SentLen; i = Chart[i]->j) {
	prin_tree_tuples(Chart[i], file);
    }
}

prin_tree_tuples (struct phrase *p, FILE *file)
{
    struct phrase *stack[1028], **kids;
    int props, top;

    stack[0] = p;
    top = 0;
    while (top >= 0) {
	p = stack[top--];
	props = phrase_props(p);
	if (GET(props, IS_VROOT)) prin_tuple(p, props, file);
	kids = p->children;
	if (kids != NULL) {
	    while (*kids != NULL) {
		if (++top >= 1028) error("prin_tree_tuples: stack overflow");
		stack[top] = *kids++;
	    }
	}
    }
}


prin_tuple (struct phrase *p, int props, FILE *file)
{
    char *hs, hp[2];
    int ispassive, cp = corpus_position(p->i);
    struct phrase *h, *f;

    h = get_headword(&hp[0], &hs, &f, p);
    ispassive = is_passive(h, f);

    if (!Pairs) fprintf(file, "%d\t%s%s", cp, hp, hs);

    prin_tuple_1(p, props, cp, hp, hs, ispassive, file);
    if (GET(props, HAS_COMPLEX_HEAD) && (p = phrase_child(p, HeadRole))) {
	props = phrase_props(p);
	prin_tuple_1(p, props, cp, hp, hs, ispassive, file);
    }
    if (!Pairs) terpri(file);
}


prin_tuple_1 (struct phrase *p, int props, int cp, char *hp, char *hs, int ispassive, FILE *file)
{
    struct phrase *a, *hd, *obj, *f, **kids;
    char *as, ap[2], *r;
    int aprops;
    
    if (GET(props, HAS_SUBJECT) && (a = phrase_child(p, SubjRole)) != NULL) {
	get_arg_head(&ap[0], &as, NULL, a);
	r = ispassive ? "obj" : "subj";
	if (Pairs) fprintf(file, "%d\t%s%s:%s\t%s%s\n", cp, hp, hs, r, ap, as);
	else fprintf(file, "\t:%s %s%s", r, ap, as);
    }
    if (GET(props, HAS_COMPLS)) {
	hd = phrase_child(p, HeadRole);
	obj = phrase_child(p, ObjRole);
	kids = p->children;
	while (*kids != NULL && *kids != hd) kids++;
	if (*kids != NULL) {
	    while ((a = *++kids) != NULL && !GET(phrase_props(a), IS_IGNORE)) {
		get_arg_head(&ap[0], &as, &f, a);
		r = a == obj ? (ispassive ? "obj2" : "obj") : compl_role(a, f);
		if (Pairs) fprintf(file, "%d\t%s%s:%s\t%s%s\n", cp, hp, hs, r, ap, as);
		else fprintf(file, "\t:%s %s%s", r, ap, as);
	    }
	}
    }
}
